from copy import deepcopy
import os
import re
from typing import Sequence, Union

import mne

from .._mne import combination_label, labels_from_mni_coords, rename_label, dissolve_label
from .definitions import DefinitionError, Definition, tuple_arg


SEEDED_PARC_RE = re.compile(r'^(.+)-(\d+)$')


class Parcellation(Definition):
    DICT_ATTRS = ('kind',)
    kind = None  # used when comparing dict representations
    morph_from_fsaverage = False

    def __init__(
            self,
            views: Union[str, Sequence[str]] = None,
    ):
        self.views = views

    def _link(self, name):
        out = deepcopy(self)
        out.name = name
        return out

    def _make(
            self,
            e,  # the MneExperiment instance
            parc: str,  # the name (contains radius for seeded parcellations)
    ) -> list:
        raise RuntimeError(f"Trying to make {self.__class__.__name__}")


class SubParc(Parcellation):
    """A subset of labels in another parcellation

    Parameters
    ----------
    base : str
        The name of the parcellation that provides the input labels. A common
        ``base`` is the ``'aparc'`` parcellation [1]_.
    labels : tuple of str
        Labels to copy from ``base``. In order to include a label in both
        hemispheres, omit the ``*-hemi`` tag. For example, with
        ``base='aparc'``, ``labels=('transversetemporal',)`` would include the
        transverse temporal gyrus in both hemisphere, whereas
        ``labels=('transversetemporal-lh',)`` would include the transverse
        temporal gyrus of only the left hemisphere.
    views : sequence of str
        Views shown in anatomical plots, e.g. ``("medial", "lateral")``.

    See Also
    --------
    MneExperiment.parcs

    Examples
    --------
    Lateral temporal lobe of both hemispheres::

        parcs = {
            'lateraltemporal': SubParc('aparc', (
                'transversetemporal', 'superiortemporal', 'bankssts',
                'middletemporal', 'inferiortemporal')),
        }

    References
    ----------
    .. [1] Desikan, R. S., Ségonne, F., Fischl, B., Quinn, B. T., Dickerson, B.
           C., Blacker, D., … Killiany, R. J. (2006). An automated labeling system
           for subdividing the human cerebral cortex on MRI scans into gyral based
           regions of interest. NeuroImage, 31(3), 968–980.
           `10.1016/j.neuroimage.2006.01.021
           <https://surfer.nmr.mgh.harvard.edu/ftp/articles/desikan06-parcellation.pdf>`_
    """
    DICT_ATTRS = ('kind', 'base', 'labels')
    kind = 'combination'

    def __init__(
            self,
            base: str,
            labels: Sequence[str],
            views: Union[str, Sequence[str]] = None,
    ):
        Parcellation.__init__(self, views)
        self.base = base
        self.labels = tuple_arg(labels)

    def _make(self, e, parc):
        with e._temporary_state:
            base = {l.name: l for l in e.load_annot(parc=self.base)}
        hemis = ('-lh', '-rh')
        labels = []
        for label in self.labels:
            if label.endswith(hemis):
                labels.append(base[label])
            else:
                for hemi in hemis:
                    labels.append(base[label + hemi])
        return labels


class CombinationParc(Parcellation):
    """Recombine labels from an existing parcellation

    Parameters
    ----------
    base
        The name of the parcellation that provides the input labels. A common
        ``base`` is the ``'aparc'`` parcellation [1]_.
    labels : dict  {str: str}
        New labels to create in ``{name: expression}`` format. All label names
        should be composed of alphanumeric characters (plus underline) and should
        not contain the -hemi tags. In order to create a given label only on one
        hemisphere, add the -hemi tag in the name (not in the expression, e.g.,
        ``{'occipitotemporal-lh': "occipital + temporal"}``).
    views
        Views shown in anatomical plots, e.g. ``("medial", "lateral")``.

    See Also
    --------
    MneExperiment.parcs

    Examples
    --------
    These are pre-defined parcellations::

        parcs = {
            'lobes-op': CombinationParc('lobes', {'occipitoparietal': "occipital + parietal"}),
            'lobes-ot': CombinationParc('lobes', {'occipitotemporal': "occipital + temporal"}),
        }

    An example using a split label. In ``split(superiorfrontal, 3)[2]``, ``3``
    indicates a split into three parts, and the index ``[2]`` picks the last
    one. Label are split along their longest axis, and ordered posterior to
    anterior, so ``[2]`` picks the most anterior part of ``superiorfrontal``::

        parcs = {
            'medial': CombinationParc('aparc', {
                'medialparietal': 'precuneus + posteriorcingulate',
                'medialfrontal': 'medialorbitofrontal + rostralanteriorcingulate'
                                 ' + split(superiorfrontal, 3)[2]',
                }, views='medial'),
        }

    References
    ----------
    .. [1] Desikan, R. S., Ségonne, F., Fischl, B., Quinn, B. T., Dickerson, B.
           C., Blacker, D., … Killiany, R. J. (2006). An automated labeling system
           for subdividing the human cerebral cortex on MRI scans into gyral based
           regions of interest. NeuroImage, 31(3), 968–980.
           `10.1016/j.neuroimage.2006.01.021
           <https://surfer.nmr.mgh.harvard.edu/ftp/articles/desikan06-parcellation.pdf>`_
    """
    DICT_ATTRS = ('kind', 'base', 'labels')
    kind = 'combination'

    def __init__(
            self,
            base: str,
            labels: dict,
            views: Union[str, Sequence[str]] = None,
    ):
        Parcellation.__init__(self, views)
        self.base = base
        self.labels = labels

    def _make(self, e, parc):
        with e._temporary_state:
            base = {l.name: l for l in e.load_annot(parc=self.base)}
        subjects_dir = e.get('mri-sdir')
        labels = []
        for name, exp in self.labels.items():
            labels += combination_label(name, exp, base, subjects_dir)
        return labels


class EelbrainParc(Parcellation):
    "Parcellation that has special make rule"
    kind = 'eelbrain_parc'

    def __init__(
            self,
            morph_from_fsaverage: bool,
            views: Union[str, Sequence[str]] = None,
    ):
        Parcellation.__init__(self, views)
        self.morph_from_fsaverage = morph_from_fsaverage

    def _make(self, e, parc):
        assert parc == 'lobes'
        subject = e.get('mrisubject')
        subjects_dir = e.get('mri-sdir')
        if subject != 'fsaverage':
            raise RuntimeError(f"lobes parcellation can only be created for fsaverage, not for {subject}")

        # load source annot
        with e._temporary_state:
            labels = e.load_annot(parc='PALS_B12_Lobes')

        # sort labels
        labels = [l for l in labels if l.name[:-3] != 'MEDIAL.WALL']

        # rename good labels
        rename_label(labels, 'LOBE.FRONTAL', 'frontal')
        rename_label(labels, 'LOBE.OCCIPITAL', 'occipital')
        rename_label(labels, 'LOBE.PARIETAL', 'parietal')
        rename_label(labels, 'LOBE.TEMPORAL', 'temporal')

        # reassign unwanted labels
        targets = ('frontal', 'occipital', 'parietal', 'temporal')
        dissolve_label(labels, 'LOBE.LIMBIC', targets, subjects_dir)
        dissolve_label(labels, 'GYRUS', targets, subjects_dir, 'rh')
        dissolve_label(labels, '???', targets, subjects_dir)
        dissolve_label(labels, '????', targets, subjects_dir, 'rh')
        dissolve_label(labels, '???????', targets, subjects_dir, 'rh')

        return labels


class FreeSurferParc(Parcellation):
    """Parcellation that is created outside Eelbrain for each subject

    Parcs that can not be generated automatically (e.g.,
    parcellation that comes with FreeSurfer). These parcellations are
    automatically scaled for brains based on scaled versions of fsaverage, but
    for individual MRIs the user is responsible for creating the respective
    annot-files.

    See Also
    --------
    MneExperiment.parcs

    Examples
    --------
    Predefined parcellations::

        parcs = {
            'aparc': FreeSurferParc(),
            }
    """
    kind = 'subject_parc'

    def _make(self, e, parc):
        subject = e.get('mrisubject')
        raise FileNotFoundError(f"At least one annot file for the parcellation {parc} is missing for {subject}")


class FSAverageParc(Parcellation):
    """Fsaverage parcellation that is morphed to individual subjects

    Parcs that are defined for the fsaverage brain and should be morphed
    to every other subject's brain. These parcellations are automatically
    morphed to individual subjects' MRIs.

    See Also
    --------
    MneExperiment.parcs

    Examples
    --------
    Predefined parcellations::

        parcs = {
            'PALS_B12_Brodmann': FSAverageParc(),
            }
    """
    kind = 'fsaverage_parc'
    morph_from_fsaverage = True

    def _make(self, e, parc):
        common_brain = e.get('common_brain')
        assert e.get('mrisubject') == common_brain
        raise FileNotFoundError(f"At least one annot file for the parcellation {parc} is missing for {common_brain}")


class LabelParc(Parcellation):
    """Assemble parcellation from FreeSurfer labels

    Combine one or several ``*.label`` files into a parcellation.

    """
    DICT_ATTRS = ('kind', 'labels')
    kind = 'label_parc'
    make = True

    def __init__(
            self,
            labels: Sequence[str],
            views: Union[str, Sequence[str]] = None,
    ):
        Parcellation.__init__(self, views)
        self.labels = tuple_arg(labels)

    def _make(self, e, parc):
        labels = []
        hemis = ('lh.', 'rh.')
        path = os.path.join(e.get('mri-dir'), 'label', '%s.label')
        for label in self.labels:
            if label.startswith(hemis):
                labels.append(mne.read_label(path % label))
            else:
                labels.extend(mne.read_label(path % (hemi + label)) for hemi in hemis)
        return labels


class SeededParc(Parcellation):
    """Parcellation that is grown from seed coordinates

    Seeds are defined on fsaverage which is in MNI305 space (`FreeSurfer wiki
    <https://surfer.nmr.mgh.harvard.edu/fswiki/CoordinateSystems>`_).
    For each seed entry, the source space vertex closest to the given coordinate
    will be used as actual seed, and a label will be created including all
    points with a surface distance smaller than a given extent from the seed
    vertex/vertices. The spatial extent is determined when setting the parc as
    analysis parameter as in ``e.set(parc="myparc-25")``, which specifies a
    radius of 25 mm.

    See Also
    --------
    MneExperiment.parcs

    Parameters
    ----------
    seeds : dict
        ``{name: seed(s)}`` dictionary, where names are strings, including
        hemisphere tags (e.g., ``"mylabel-lh"``) and seed(s) are array-like,
        specifying one or more seed coordinate (shape ``(3,)`` or
        ``(n_seeds, 3)``).
    mask : str
        Name of a parcellation to use as mask (i.e., anything that is "unknown"
        in that parcellation is excluded from the new parcellation. For example,
        use ``{'mask': 'lobes'}`` to exclude the subcortical areas around the
        diencephalon.

    Examples
    --------
    Example with multiple seeds::

         parcs = {
             'stg': SeededParc({
                 'anteriorstg-lh': ((-54, 10, -8), (-47, 14, -28)),
                 'middlestg-lh': (-66, -24, 8),
                 'posteriorstg-lh': (-54, -57, 16),
             },
             mask='lobes'),
         }
    """
    DICT_ATTRS = ('kind', 'seeds', 'surface', 'mask')
    kind = 'seeded'
    make = True

    def __init__(self, seeds, mask=None, surface='white', views=None):
        Parcellation.__init__(self, views)
        self.seeds = seeds
        self.mask = mask
        self.surface = surface

    def seeds_for_subject(self, subject):
        return self.seeds

    def _make(self, e, parc):
        if self.mask:
            with e._temporary_state:
                e.make_annot(parc=self.mask)
        subject = e.get('mrisubject')
        subjects_dir = e.get('mri-sdir')
        seeds = self.seeds_for_subject(subject)
        name, extent = SEEDED_PARC_RE.match(parc).groups()
        return labels_from_mni_coords(seeds, float(extent), subject, self.surface, self.mask, subjects_dir, parc)


class IndividualSeededParc(SeededParc):
    """Seed parcellation with individual seeds for each subject

    Analogous to :class:`SeededParc`, except that seeds are
    provided for each subject.

    See Also
    --------
    MneExperiment.parcs

    Examples
    --------
    Parcellation with subject-specific seeds::

        parcs = {
            'stg': IndividualSeededParc({
                'anteriorstg-lh': {
                    'R0001': (-54, 10, -8),
                    'R0002': (-47, 14, -28),
                },
                'middlestg-lh': {
                    'R0001': (-66, -24, 8),
                    'R0002': (-60, -26, 9),
                }
                mask='lobes'),
        }
    """
    kind = 'individual seeded'
    morph_from_fsaverage = False

    def __init__(self, seeds, mask=None, surface='white', views=None):
        SeededParc.__init__(self, seeds, mask, surface, views)
        labels = tuple(self.seeds)
        label_subjects = {label: sorted(self.seeds[label].keys()) for label in labels}
        subjects = label_subjects[labels[0]]
        if not all(label_subjects[label] == subjects for label in labels[1:]):
            raise DefinitionError("Some labels are missing subjects")
        self.subjects = subjects

    def seeds_for_subject(self, subject):
        if subject not in self.subjects:
            raise DefinitionError(f"Parcellation {self.name} not defined for subject {subject}")
        seeds = {name: self.seeds[name][subject] for name in self.seeds}
        # filter out missing
        return {name: seed for name, seed in seeds.items() if seed}


def parc_from_dict(name, params):
    p = params.copy()
    kind = p.pop('kind', None)
    if kind is None:
        raise KeyError(f"Parcellation {name} does not contain the required 'kind' entry")
    elif kind not in PARC_CLASSES:
        raise ValueError(f"Parcellation {name} contains an invalid 'kind' entry: {kind!r}")
    cls = PARC_CLASSES[kind]
    return cls(**p)


PARC_CLASSES = {p.kind: p for p in (CombinationParc, FreeSurferParc, FSAverageParc, SeededParc, IndividualSeededParc)}


def assemble_parcs(items):
    parcs = {}
    for name, obj in items:
        if isinstance(obj, Parcellation):
            parc = obj
        elif isinstance(obj, dict):
            parc = parc_from_dict(name, obj)
        elif obj == FreeSurferParc.kind:
            parc = FreeSurferParc(('lateral', 'medial'))
        elif obj == FSAverageParc.kind:
            parc = FSAverageParc(('lateral', 'medial'))
        else:
            raise DefinitionError(f"parcellation {name!r}: {obj!r}")
        parcs[name] = parc._link(name)
    return parcs



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Plot data split for cross-validation"""
from typing import Union

from matplotlib.patches import Rectangle

from .._trf.shared import Splits
from ._base import EelFigure, Layout, LegendMixin
from ._styles import colors_for_oneway


class DataSplit(EelFigure, LegendMixin):

    def __init__(
            self,
            splits: Splits,
            legend: Union[str, int, bool] = 'upper right',
            colors: dict = None,
            **kwargs,
    ):
        """Plot data splits

        Parameters
        ----------
        splits
            The data splits.
        legend
            Matplotlib figure legend location argument or 'fig' to plot the
            legend in a separate figure.
        colors
            Customize colors (uses the following keys: ``'train', 'validate',
            'test'``).
        ...
            Also accepts :ref:`general-layout-parameters`.

        Notes
        -----
        This plot can also be generated from a :class:`BoostingResult` object
        through :meth:`BoostingResult.splits.plot` method.
        """
        attrs = ['train', 'validate', 'test']
        labels = {'train': 'Training', 'validate': 'Validation', 'test': 'Testing'}
        if colors is None:
            colors = colors_for_oneway(attrs, unambiguous=[6, 3, 5])

        layout = Layout(1, 16/9, 2, **kwargs)
        EelFigure.__init__(self, None, layout)
        ax = self.figure.axes[0]

        for x in splits.segments[:-1, 1]:
            ax.axvline(x, color='k')
        handles = {}
        for y, split in enumerate(splits.splits):
            for attr in attrs:
                segments = getattr(split, attr)
                if segments is None:
                    continue
                rect = None
                for x0, x1 in segments:
                    rect = Rectangle((x0, y-0.5), x1-x0, 1, color=colors[attr])
                    ax.add_artist(rect)
                handles[attr] = rect
            # separate splits
            if y:
                ax.axhline(y-0.5, color='white')
        labels = {key: labels[key] for key in handles}
        ax.set_ylabel('Split')
        ax.set_ylim(-0.5, len(splits.splits)-0.5)
        ax.set_xlabel('Sample')
        ax.set_xlim(splits.segments[0, 0], splits.segments[-1, 1])
        LegendMixin.__init__(self, legend, handles, labels)
        self._show()

    def _fill_toolbar(self, tb):
        LegendMixin._fill_toolbar(self, tb)



from collections.abc import Iterator
from dataclasses import dataclass, fields
from functools import reduce
from itertools import product
from operator import mul
from typing import List, Union

import numpy as np
import scipy.signal
from scipy.linalg import norm

from .. import _info
from .._data_obj import CategorialArg, NDVarArg, Dataset, NDVar, Case, UTS, dataobj_repr, ascategorial, asndvar
from .._utils import LazyProperty, PickleableDataClass
from .._utils.numpy_utils import newaxis


class EQMixIn:

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        for field in fields(self):
            a, b = getattr(self, field.name), getattr(other, field.name)
            if isinstance(a, b.__class__):
                if isinstance(a, np.ndarray):
                    if not np.array_equal(a, b):
                        break
                elif a != b:
                    break
            else:
                break
        else:
            return True
        return False


@dataclass(eq=False)
class Split(PickleableDataClass, EQMixIn):
    train: np.ndarray  # (, 2) array of int, segment (start, stop)
    validate: np.ndarray = None
    test: np.ndarray = None
    i_test: int = 0  # Index (to group splits with the same test segmet)

    @LazyProperty
    def train_and_validate(self):
        return np.vstack([self.train, self.validate])


def merge_segments(
        segments: np.ndarray,
        soft_splits: Union[bool, np.ndarray] = None,
):
    """Take a selection of input segments and remove soft boundaries"""
    # return out_segments
    if soft_splits is None or isinstance(soft_splits, np.ndarray) and len(soft_splits) == 0:
        return segments
    out_segments = list(segments)
    for i in range(len(out_segments)-1, 0, -1):
        pre_seg, post_seg = out_segments[i-1], out_segments[i]
        if pre_seg[1] >= post_seg[0]:
            if soft_splits is True or post_seg[0] in soft_splits:
                out_segments[i-1] = [pre_seg[0], max(pre_seg[1], post_seg[1])]
                del out_segments[i]
    return np.vstack(out_segments)


@dataclass(eq=False)
class Splits(PickleableDataClass, EQMixIn):
    splits: List[Split]
    partitions_arg: Union[int, None]
    n_partitions: int
    n_validate: int
    n_test: int
    model: CategorialArg = None
    segments: np.ndarray = None  # Original data segments
    split_segments: np.ndarray = None  # Data subdivision for splits

    def __repr__(self):
        if len(self.segments) == 1:
            desc = "continuous data"
        else:
            desc = f"{len(self.segments)} data segments"
        items = ['']
        if self.n_validate:
            items.append(f'n_validate={self.n_validate}')
        if self.n_test:
            items.append(f'n_test={self.n_test}')
        if self.model is not None:
            items.append(f'model={dataobj_repr(self.model)}')
        items = ', '.join(items)
        return f"<Splits: {desc} split into {len(self.split_segments)} sections{items}>"

    def plot(self, **kwargs):
        """Plot data splits (see :class:`plot.SplitFigure` for parameters)"""
        from ..plot import DataSplit

        return DataSplit(self, **kwargs)


def split_data(
        segments: np.ndarray,  # (n, 2) array of [start, stop] indices
        partitions: int = None,  # Number of segments to split the data
        model: CategorialArg = None,  # sample evenly from cells
        ds: Dataset = None,
        validate: int = 1,  # Number of segments in validation set
        test: int = 0,  # Number of segments in test set
):
    """Split data segments into train, validate and test segments"""
    if partitions is not None and partitions <= validate + test:
        raise ValueError(f"partitions={partitions} with validate={validate}, test={test}")
    partitions_arg = partitions
    assert validate >= 0
    if validate > 1:
        raise NotImplementedError
    assert test >= 0
    if test > 1:
        raise NotImplementedError
    if len(segments) == 1:
        if partitions is None:
            partitions = 2 + test + validate if test else 10
        if model is not None:
            raise TypeError(f'model={dataobj_repr(model)!r}: model cannot be specified in unsegmented data')
        n_times = segments[0, 1] - segments[0, 0]
        split_points = np.round(np.linspace(0, n_times, partitions + 1)).astype(np.int64)
        soft_splits = split_points[1:-1]
        split_segments = np.vstack([split_points[i: i+2] for i in range(partitions)])
        categories = [None]
    else:
        n_segments = len(segments)
        # determine model cells
        if model is None:
            categories = [None]
            cell_size = n_segments
        else:
            model = ascategorial(model, ds=ds, n=n_segments)
            categories = [np.flatnonzero(model == cell) for cell in model.cells]
            cell_sizes = [len(i) for i in categories]
            cell_size = min(cell_sizes)
            cell_sizes_are_equal = len(set(cell_sizes)) == 1
            if partitions is None and not cell_sizes_are_equal:
                raise NotImplementedError(f'Automatic partition for variable cell size {dict(zip(model.cells, cell_sizes))}')
        # automatic selection of partitions
        if partitions is None:
            if 3 <= cell_size <= 10:
                partitions = cell_size
            else:
                raise NotImplementedError(f"Automatic partition for {cell_size} cases")
        # create segments
        if cell_size >= partitions:
            soft_splits = None
            split_segments = segments
        else:
            # need to subdivide segments
            if model is not None:
                raise NotImplementedError(f'partitions={partitions}: with model')
            elif partitions % cell_size:
                raise ValueError(f'partitions={partitions}: not a multiple of n_cases ({cell_size})')
            n_parts = partitions // cell_size
            split_segments = []
            soft_splits = []
            for start, stop in segments:
                split_points = np.round(np.linspace(start, stop, n_parts)).astype(np.int64)
                soft_splits.append(split_points[1:-1])
                split_segments.extend(split_points[i: i + 2] for i in range(n_parts))
            soft_splits = np.concatenate(soft_splits)
            split_segments = np.vstack(split_segments)
    # create actual splits
    splits = []  # list of Split
    test_iter = range(partitions) if test else [None]
    validate_iter = range(partitions) if validate else [None]
    n_segments = len(split_segments)
    for i_test in test_iter:
        for i_validate in validate_iter:
            if i_test == i_validate:
                continue
            train_set = np.ones(n_segments, bool)
            # test set
            if i_test is None:
                test_segments = None
            else:
                test_set = np.zeros(n_segments, bool)
                for cell_index in categories:
                    index = slice(i_test, None, partitions)
                    if cell_index is not None:
                        index = cell_index[index]
                    test_set[index] = True
                train_set ^= test_set
                test_segments = merge_segments(split_segments[test_set], soft_splits)
            # validation set
            if i_validate is None:
                validate_segments = None
            else:
                validate_set = np.zeros(n_segments, bool)
                for cell_index in categories:
                    index = slice(i_validate, None, partitions)
                    if cell_index is not None:
                        index = cell_index[index]
                    validate_set[index] = True
                train_set ^= validate_set
                validate_segments = merge_segments(split_segments[validate_set], soft_splits)
            # create split
            train_segments = merge_segments(split_segments[train_set], soft_splits)
            splits.append(Split(train_segments, validate_segments, test_segments, i_test))
    return Splits(splits, partitions_arg, partitions, validate, test, model, segments, split_segments)


class PredictorData:
    """Restructure model NDVars (like RevCorrData but for x only)"""

    def __init__(
            self,
            x: Union[NDVarArg, List[NDVarArg]],
            ds: Dataset = None,
    ):
        if isinstance(x, (tuple, list, Iterator)):
            multiple_x = True
            xs = [asndvar(x_, ds=ds) for x_ in x]
            if len(xs) == 0:
                raise ValueError(f"x={x!r} of length 0")
            x_name = [x_.name for x_ in xs]
        else:
            multiple_x = False
            xs = [asndvar(x, ds=ds)]
            x_name = xs[0].name

        time_dim = xs[0].get_dim('time')
        if any(xi.get_dim('time') != time_dim for xi in xs[1:]):
            raise ValueError("Not all NDVars in x have the same time dimension")
        n_times = len(time_dim)

        # determine cases (used as segments)
        has_case = n_cases = segments = None
        for xi in xs:
            # determine cases
            if n_cases is None:
                has_case = xi.has_case
                if xi.has_case:
                    n_cases = len(xi)
                    # prepare segment index
                    seg_i = np.arange(0, n_cases * n_times + 1, n_times, np.int64)[:, newaxis]
                    segments = np.hstack((seg_i[:-1], seg_i[1:]))
                else:
                    n_cases = 0
                    segments = np.array([[0, n_times]], np.int64)
            elif xi.has_case ^ has_case:
                raise ValueError(f'x={xs}: some but not all x have case')
            elif has_case and len(xi) != n_cases:
                raise ValueError(f'x={xs}: not all items have the same number of cases')

        # x_data:  predictor x time array
        x_data = []
        x_meta = []
        x_names = []
        n_x = 0
        for xi in xs:
            ndim = xi.ndim - bool(n_cases)
            if has_case:
                dimnames = xi.get_dimnames(last=('case', 'time'))
                xdims = xi.get_dims(dimnames[:-2])
                shape = (-1, n_cases * n_times)
            else:
                dimnames = xi.get_dimnames(last='time')
                xdims = xi.get_dims(dimnames[:-1])
                shape = (-1, n_times)
            data = xi.get_data(dimnames).reshape(shape)
            x_repr = dataobj_repr(xi)
            if ndim == 1:
                index = n_x
                x_names.append(x_repr)
            else:
                index = slice(n_x, n_x + len(data))
                for v in product(*xdims):
                    x_names.append("-".join((x_repr, *map(str, v))))
            x_data.append(data)
            x_meta.append((xi.name, xdims, index))
            n_x += len(data)

        if len(x_data) == 1:
            x_data = x_data[0]
            x_data_is_copy = False
        else:
            x_data = np.concatenate(x_data)
            x_data_is_copy = True

        self.has_case = has_case
        self.n_cases = n_cases
        self.case_to_segments = n_cases > 0
        self.time_dim = time_dim
        self.n_times = n_times
        self.n_times_flat = n_cases * n_times if n_cases else n_times
        self.multiple_x = multiple_x
        self.x_name = x_name
        self.x_names = x_names
        self.x_meta = x_meta
        self.data = x_data
        self.data_is_copy = x_data_is_copy
        self.segments = segments


class RevCorrData:
    """Restructure input NDVars into arrays for reverse correlation
    
    Attributes
    ----------
    y : NDVar
        Dependent variable.
    x : NDVar | sequence of NDVar
        Predictors.
    segments : np.ndarray
        ``(n_segments, 2)`` array of segment ``[start, stop]`` indices. The
        segments delimit chunks of continuous data, such as trials.
    splits : list of Split
        Cross-validation scheme.
    """
    # data
    x_mean = None
    x_scale = None
    y_mean = None
    y_scale = None
    x_pads = None
    _x_is_copy: bool = False
    _y_is_copy: bool = False
    scale_data: str = None
    # cross-validation
    splits: Splits = None

    def __init__(
            self,
            y: NDVarArg,
            x: Union[NDVarArg, List[NDVarArg]],
            ds: Dataset = None,
            in_place: bool = False,
    ):
        x_data = PredictorData(x, ds)

        # check y
        y = asndvar(y, ds=ds)
        if y.get_dim('time') != x_data.time_dim:
            raise ValueError("y does not have the same time dimension as x")
        if y.has_case ^ x_data.has_case:
            raise ValueError(f'{y=}: case dimension does not match x')
        elif y.has_case and len(y) != x_data.n_cases:
            raise ValueError(f'{y=}: different number of cases from x ({x_data.n_cases})')

        # vector dimension
        vector_dims = [dim.name for dim in y.dims if dim._connectivity_type == 'vector']
        if not vector_dims:
            vector_dim = None
        elif len(vector_dims) == 1:
            vector_dim = y.get_dim(vector_dims.pop())
        else:
            raise NotImplementedError(f"{y=}: more than one vector dimension ({', '.join(vector_dims)})")

        # y_data: flatten to ydim x time array
        last = ('time',)
        n_ydims = -1
        if x_data.case_to_segments:
            last = ('case', *last)
            n_ydims -= 1
        if vector_dim:
            last = (vector_dim.name, *last)
        y_dimnames = y.get_dimnames(last=last)
        ydims = y.get_dims(y_dimnames[:n_ydims])
        n_flat = reduce(mul, map(len, ydims), 1)
        shape = (n_flat, x_data.n_times_flat)
        y_data = y.get_data(y_dimnames).reshape(shape)
        # shape for exposing vector dimension
        if vector_dim:
            n_flat_prevector = reduce(mul, map(len, ydims[:-1]), 1)
            n_vector = len(ydims[-1])
            assert n_vector > 1
            vector_shape = (n_flat_prevector, n_vector, x_data.n_times_flat)
        else:
            vector_shape = None

        self.time = x_data.time_dim
        self.segments = x_data.segments
        self.shortest_segment_n_times = x_data.n_times
        self.in_place = in_place
        # y
        self.y = y_data  # (n_signals, n_times)
        self.y_name = y.name
        self._y_repr = dataobj_repr(y)
        self.y_info = _info.copy(y.info)
        self.ydims = ydims  # without case and time
        self.yshape = tuple(map(len, ydims))
        self.full_y_dims = y.get_dims(y_dimnames)
        self.vector_dim = vector_dim  # vector dimension
        self.vector_shape = vector_shape  # flat shape with vector dim separate
        # x
        self.x = x_data.data  # (n_predictors, n_times)
        self.x_name = x_data.x_name
        self.x_names = x_data.x_names
        self._x_meta = x_data.x_meta  # [(x.name, xdim, index), ...]; index is int or slice
        self._multiple_x = x_data.multiple_x
        self._x_is_copy = x_data.data_is_copy
        # basis
        self.basis = 0
        self.basis_window = None

    def _copy_data(self, y=False):
        "Make sure the data is a copy before modifying"
        if self.in_place:
            return
        if not self._x_is_copy:
            self.x = self.x.copy()
            self._x_is_copy = True
        if y and not self._y_is_copy:
            self.y = self.y.copy()
            self._y_is_copy = True

    def apply_basis(self, basis: float, basis_window: str):
        """Apply basis to x

        Notes
        -----
        Normalize after applying basis (basis can smooth out variance).
        """
        if self.basis != 0:
            raise NotImplementedError("Applying basis more than once")
        elif not basis:
            return
        self._copy_data()
        n = int(round(basis / self.time.tstep))
        w = scipy.signal.get_window(basis_window, n, False)
        if len(w) <= 1:
            raise ValueError(f"basis={basis!r}: Window is {len(w)} samples long")
        w /= w.sum()
        for xi in self.x:
            xi[:] = scipy.signal.convolve(xi, w, 'same')
        self.basis = basis
        self.basis_window = basis_window

    @LazyProperty
    def x_pads(self):
        return np.zeros(len(self.x))

    def normalize(self, error: str):
        self._copy_data(y=True)
        y_mean = self.y.mean(1)
        x_mean = self.x.mean(1)
        self.y -= y_mean[:, newaxis]
        self.x -= x_mean[:, newaxis]
        # for vector data, scale by vector norm
        if self.vector_shape:
            y_data_vector_shape = self.y.reshape(self.vector_shape)
            y_data_for_scale = norm(y_data_vector_shape, axis=1)
        else:
            y_data_vector_shape = None
            y_data_for_scale = self.y

        if error == 'l1':
            y_scale = np.abs(y_data_for_scale).mean(-1)
            x_scale = np.abs(self.x).mean(-1)
        elif error == 'l2':
            y_scale = (y_data_for_scale ** 2).mean(-1) ** 0.5
            x_scale = (self.x ** 2).mean(-1) ** 0.5
        else:
            raise RuntimeError(f"error={error!r}")

        if self.vector_shape:
            y_data_vector_shape /= y_scale[:, newaxis, newaxis]
        else:
            self.y /= y_scale[:, newaxis]
        self.x /= x_scale[:, newaxis]

        self.scale_data = error
        self.y_mean = y_mean
        self.y_scale = y_scale
        self.x_mean = x_mean
        self.x_scale = x_scale
        # zero-padding for convolution
        self.x_pads = -x_mean / x_scale

    def _check_data(self):
        if self.x_scale is None:
            x_check = self.x.var(1)
            y_check = self.y.var(1)
        else:
            x_check = self.x_scale
            y_check = self.y_scale
        # check for flat data
        zero_var = y_check == 0
        if np.any(zero_var):
            raise ValueError(f"y={self._y_repr}: contains {zero_var.sum()} flat time series")
        zero_var = x_check == 0
        if np.any(zero_var):
            names = [self.x_names[i] for i in np.flatnonzero(zero_var)]
            raise ValueError(f"x: flat data in {', '.join(names)}")
        # check for NaN
        has_nan = np.isnan(y_check.sum())
        if has_nan:
            raise ValueError(f"y={self._y_repr}: contains NaN")
        has_nan = np.isnan(x_check)
        if np.any(has_nan):
            names = [self.x_names[i] for i in np.flatnonzero(has_nan)]
            raise ValueError(f"x: NaN in {', '.join(names)}")

    def initialize_cross_validation(
            self,
            partitions: int = None,  # Number of segments to split the data
            model: CategorialArg = None,  # sample evenly from cells
            ds: Dataset = None,
            validate: int = 1,  # Number of segments in validation set
            test: int = 0,  # Number of segments in test set
    ):
        """Initialize cross-validation scheme

        Notes
        -----
        General solution:

         - split data into even sized segments (hard and soft splits)
         - group segments by cell-index
         - create splits
         - merge segments at soft boundaries
        """
        self.splits = split_data(self.segments, partitions, model, ds, validate, test)

    def data_scale_ndvars(self):
        if self.scale_data:
            # y
            if self.yshape:
                y_mean = NDVar(self.y_mean.reshape(self.yshape), self.ydims, self.y_name, self.y_info)
            else:
                y_mean = self.y_mean[0]
            # scale does not include vector dim
            if self.vector_dim:
                dims = self.ydims[:-1]
                shape = self.yshape[:-1]
            else:
                dims = self.ydims
                shape = self.yshape
            if shape:
                y_scale = NDVar(self.y_scale.reshape(shape), dims, self.y_name, self.y_info)
            else:
                y_scale = self.y_scale[0]
            # x
            x_mean = []
            x_scale = []
            for name, xdims, index in self._x_meta:
                if xdims:
                    shape = [len(dim) for dim in xdims]
                    x_mean.append(NDVar(self.x_mean[index].reshape(shape), xdims, name))
                    x_scale.append(NDVar(self.x_scale[index].reshape(shape), xdims, name))
                else:
                    x_mean.append(self.x_mean[index])
                    x_scale.append(self.x_scale[index])
            if self._multiple_x:
                x_mean = tuple(x_mean)
                x_scale = tuple(x_scale)
            else:
                x_mean = x_mean[0]
                x_scale = x_scale[0]
        else:
            y_mean = y_scale = x_mean = x_scale = None
        return y_mean, y_scale, x_mean, x_scale

    def package_kernel(self, h, tstart):
        """Package kernel as NDVar
        
        Parameters
        ----------
        h : array  (n_y, n_x, n_times)
            Kernel data.
        """
        h_time = UTS(tstart, self.time.tstep, h.shape[-1], self.time.unit)
        hs = []
        if self.scale_data:
            info = _info.for_normalized_data(self.y_info, 'Response')
        else:
            info = self.y_info

        for name, xdims, index in self._x_meta:
            dims = (*self.ydims, *xdims, h_time)
            shape = [len(dim) for dim in dims]
            x = h[:, index, :].reshape(shape)
            hs.append(NDVar(x, dims, name, info))

        if self._multiple_x:
            return tuple(hs)
        else:
            return hs[0]

    def package_value(
            self,
            value: np.ndarray,  # data
            name: str,  # NDVar name
            info: dict = None,  # NDVar info
            meas: str = None,  # for NDVar info
    ):
        if not self.yshape:
            return value[0]

        # shape
        has_vector = value.shape[0] > self.yshape[0]
        if self.vector_dim and not has_vector:
            dims = self.ydims[:-1]
            shape = self.yshape[:-1]
        else:
            dims = self.ydims
            shape = self.yshape
        if not dims:
            return value[0]
        elif len(shape) > 1:
            value = value.reshape(shape)

        # info
        if meas:
            info = _info.for_stat_map(meas, old=info)
        elif info is None:
            info = self.y_info

        return NDVar(value, dims, name, info)

    def package_y_like(self, data, name):
        shape = tuple(map(len, self.full_y_dims))
        data = data.reshape(shape)
        # roll Case to first axis
        for axis, dim in enumerate(self.full_y_dims):
            if isinstance(dim, Case):
                data = np.rollaxis(data, axis)
                dims = list(self.full_y_dims)
                dims.insert(0, dims.pop(axis))
                break
        else:
            dims = self.full_y_dims
        return NDVar(data, dims, name)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from eelbrain._utils.parse import find_variables


def test_find_variables():
    assert find_variables("a + b / c.x()") == {'a', 'b', 'c'}
    assert find_variables("a + 'b' / c.x()") == {'a', 'c'}



"""Eelbrain GUIs"""


def run(block=False):
    """Hand over command to the GUI (quit the GUI to return to the terminal)

    Parameters
    ----------
    block : bool
        Block the Terminal even if the GUI is capable of being run in parallel.
        Control returns to the Terminal when the user quits the GUI application.
        This is also useful to prevent plots from closing at the end of a
        script.
    """
    from ._wxgui import run

    run(block)


def select_components(path, ds, sysname=None, connectivity=None):
    """GUI for selecting ICA-components

    Parameters
    ----------
    path : str
        Path to the ICA file.
    ds : Dataset
        Dataset with epochs to use for source selection in ``ds['epochs']``
        (as mne-python ``Epochs`` object). Optionally, ``ds['index']`` can be
        the indexes to display for epochs (the default is ``range(n_epochs)``.
        Further :class:`Factor` can be used to plot condition averages.
    sysname : str
        Optional, to define sensor connectivity.
    connectivity :
        Optional, to define sensor connectivity (see
        :func:`eelbrain.load.fiff.sensor_dim`).

    Notes
    -----
    The ICA object does not need to be computed on the same data that is in
    ``ds``. For example, the ICA can be computed on a raw file but component
    selection done using the epochs that will be analyzed.
    """
    from ._wxgui.app import get_app
    from ._wxgui.select_components import TEST_MODE, Document, Frame, Model

    get_app()  # make sure app is created
    doc = Document(path, ds, sysname, connectivity)
    model = Model(doc)
    frame = Frame(None, None, None, model)
    frame.Show()
    frame.Raise()
    if TEST_MODE:
        return frame
    else:
        run()


def select_epochs(
        ds, data='meg', accept='accept', blink='blink', tag='rej_tag',
        trigger='trigger', path=None, nplots=None, topo=None, mean=None,
        vlim=None, color='k', lw=0.5, mark=[], mcolor='r', mlw=0.8,
        antialiased=True, pos=None, size=None, allow_interpolation=True):
    """GUI for rejecting trials of MEG/EEG data

    Parameters
    ----------
    ds : Dataset | mne.Epochs
        The data for which to select trials. If ``ds`` is a :class:`Dataset`,
        the subsequent parameters up to 'trigger' specify names of variables in
        ``ds``; if ``ds`` is an :class:`mne.Epochs` object, these parameters are
        ignored.
    data : str
        Name of the epochs data in ``ds`` (default ``'meg'``).
    accept : str
        Name of the boolean :class:`Var` in ``ds`` to accept or reject epochs
        (default ``'accept'``).
    blink : str
        Name of the eye tracker data in ``ds`` if present (default ``'blink'``).
    tag : str
        Name of the rejection tag (storing the reason for rejecting a
        specific epoch; default ``'rej_tag'``).
    trigger : str
        Name of the ``int`` :class:`Var` containing the event trigger for each
        epoch (used when loading a rejection file to assert that it comes from
        the same data; default ``'trigger'``).
    path : None | str
        Path to the desired rejection file. If the file already exists it
        is loaded as starting values. The extension determines the format
        (``*.pickled`` or ``*.txt``).
    nplots : None | int | tuple of 2 int
        Number of epoch plots per page. Can be an ``int`` to produce a
        square layout with that many epochs, or an ``(n_rows, n_columns)``
        tuple. The default is to use the settings from the last session.
    topo : None | bool
        Show a topomap plot of the time point under the mouse cursor.
        Default (None): use settings form last session.
    mean : None | bool
        Show a plot of the page mean at the bottom right of the page.
        Default (None): use settings form last session.
    vlim : None | scalar
        Limit of the epoch plots on the y-axis. If None, a value is
        determined automatically to show all data.
    color : None | matplotlib color
        Color for primary data (default is black).
    lw : scalar
        Linewidth for normal sensor plots.
    mark : None | index for sensor dim
        Sensors to plot as individual traces with a separate color.
    mcolor : matplotlib color
        Color for marked traces.
    mlw : scalar
        Line width for marked sensor plots.
    antialiased : bool
        Perform Antialiasing on epoch plots (associated with a minor speed
        cost).
    pos : None | tuple of 2 int
        Window position on screen. Default (None): use settings form last
        session.
    size : None | tuple of 2 int
        Window size on screen. Default (None): use settings form last
        session.
    allow_interpolation : bool
        Whether to allow interpolating individual channels by epoch (default
        True).

    Notes
    -----
    Exclude bad epochs and interpolate or remove bad channels.

    * Use the `Bad Channels` button in the toolbar to exclude channels from
      analysis (use the `GA` button to plot the grand average and look for
      channels that are consistently bad).
    * Click the `Threshold` button to automatically reject epochs in which the
      signal exceeds a certain threshold.
    * Click on an epoch plot to toggle rejection of that epoch.
    * Press ``i`` on the keyboard to toggle channel interpolation for the
      channel that is closest to the cursor along the y-axis.
    * Press ``shift-i`` on the keyboard to edit a list of interpolated channels
      for the epoch under the cursor.


    *Keyboard shortcuts* in addition to the ones in the menu:

    =========== ============================================================
    Key         Effect
    =========== ============================================================
    right-arrow go to the next page
    left-arrow  go to the previous page
    b           butterfly plot of the epoch under the pointer
    c           pairwise sensor correlation plot or the current epoch
    t           topomap plot of the epoch/time point under the pointer
    i           interpolate the channel nearest to the pointer on the y-axis
    shift-i     open dialog to enter channels for interpolation
    =========== ============================================================
    """
    from ._wxgui.app import get_app
    from ._wxgui.select_epochs import TEST_MODE, Document, Frame, Model

    get_app()  # make sure app is created
    bad_chs = None
    doc = Document(ds, data, accept, blink, tag, trigger, path, bad_chs, allow_interpolation)
    model = Model(doc)
    frame = Frame(None, model, nplots, topo, mean, vlim, color, lw, mark, mcolor, mlw, antialiased, pos, size, allow_interpolation)
    frame.Show()
    frame.Raise()
    if TEST_MODE:
        return frame
    else:
        run()


def load_stcs():
    """GUI for detecting and loading source estimates

    Notes
    -----
    This GUI prompts the user for the directory containing the
    source estimates, then automatically detects the experimental
    design, and loads the data into a :class:`Dataset`. The dataset
    can be added to the IPython session's namespace, or passed on
    to the statistics GUI.

    See Also
    --------
    eelbrain.load.fiff.DatasetSTCLoader : programmatic access
        to this functionality
    """
    from ._wxgui.app import get_app
    from ._wxgui.load_stcs import STCLoaderFrame, TEST_MODE

    get_app()
    frame = STCLoaderFrame(None)
    if TEST_MODE:
        return frame
    else:
        run()



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from inspect import getfullargspec
import re
from typing import Collection, Tuple, Union

from .. import testnd
from .. import test
from .._data_obj import CellArg
from .._exceptions import DefinitionError
from .._io.fiff import find_mne_channel_types
from .._utils.parse import find_variables
from .definitions import Definition
from .variable_def import Variables, GroupVar


__test__ = False
TAIL_REPR = {0: '=', 1: '>', -1: '<'}


def assemble_tests(test_dict):
    "Interpret dict with test definitions"
    out = {}
    for key, params in test_dict.items():
        if isinstance(params, Test):
            out[key] = params
            continue
        elif not isinstance(params, dict):
            raise TypeError(f"Invalid object for test definition {key}: {params!r}")
        params = params.copy()
        if 'stage 1' in params:
            params['stage_1'] = params.pop('stage 1')
        kind = params.pop('kind')
        if kind in TEST_CLASSES:
            out[key] = TEST_CLASSES[kind](**params)
        else:
            raise DefinitionError(f"Unknown test kind in test definition {key}: {kind}")
    return out


def tail_arg(tail):
    try:
        if tail == 0:
            return 0
        elif tail > 0:
            return 1
        else:
            return -1
    except Exception:
        raise TypeError("tail=%r; needs to be 0, -1 or 1" % (tail,))


class Test(Definition):
    "Baseclass for any test"
    kind = None
    DICT_ATTRS = ('kind', 'model', 'vars')

    def __init__(
            self,
            desc: str,
            model: str = None,  # within-subject model; None for single-trial analysis
            vars: Union[str, tuple, list, dict] = None,  # dynamic variables
            cat: Tuple[CellArg, ...] = None,  # cells in model to load
            depend_on: Collection[str] = (),  # non-model variables
    ):
        self.desc = desc
        if model is None:
            self._test_vars = []
            self.model = None
        else:
            self._test_vars = [v for v in map(str.strip, model.split('%')) if v]
            self.model = '%'.join(self._test_vars)
        self.cat = cat
        try:
            self.vars = Variables(vars)
        except Exception as error:
            raise DefinitionError(f"vars={vars} ({error})")
        self._test_vars.extend(depend_on)

    def _find_test_vars(self):
        "Find variables and groups used in a test definition"
        vs = set(self._test_vars)
        groups = set()
        for name, variable in self.vars.vars.items():
            if name in vs:
                vs.remove(name)
                vs.update(variable.input_vars())
                if isinstance(variable, GroupVar):
                    groups.update(variable.groups)
        return vs, groups

    def make(self, y, ds, force_permutation, kwargs):
        raise NotImplementedError(f"For {self.__class__.__name__}")

    def make_uv(self, y, ds):
        raise NotImplementedError(f"For {self.__class__.__name__}")


class TTestOneSample(Test):
    """One-sample t-test

    Parameters
    ----------
    tail : int
        Tailedness of the test. ``0`` for two-tailed (default), ``1`` for upper tail
        and ``-1`` for lower tail.

    See Also
    --------
    MneExperiment.tests
    """
    kind = 'ttest_1samp'
    DICT_ATTRS = Test.DICT_ATTRS + ('tail',)

    def __init__(self, tail: int = 0):
        tail = tail_arg(tail)
        desc = "%s 0" % TAIL_REPR[tail]
        Test.__init__(self, desc, '')
        self.tail = tail

    def make(self, y, ds, force_permutation, kwargs):
        return testnd.TTestOneSample(y, match='subject', ds=ds, tail=self.tail, force_permutation=force_permutation, **kwargs)

    def make_uv(self, y, ds):
        return test.TTestOneSample(y, match='subject', ds=ds, tail=self.tail)


class TTestIndependent(Test):
    """Independent measures t-test (comparing groups of subjects)

    Parameters
    ----------
    model : str
        The model which defines the cells that are used in the test. Usually
        ``"group"``.
    c1 : str | tuple
        The experimental group. Should be a group name.
    c0 : str | tuple
        The control group, defined like ``c1``.
    tail : int
        Tailedness of the test. ``0`` for two-tailed (default), ``1`` for upper tail
        and ``-1`` for lower tail.

    See Also
    --------
    MneExperiment.tests

    Examples
    --------
    Sample test definitions, assuming that the experiment has two groups called
    ``'younger'`` and ``'older'``::

        variables = {
            'age': GroupVar(['younger', 'older']),
        }
        tests = {
            'old=young': TTestIndependent('group', 'older', 'younger'),
            'old>young': TTestIndependent('group', 'older', 'younger', tail=1),
        }
    """
    kind = 'ttest_ind'
    DICT_ATTRS = Test.DICT_ATTRS + ('c1', 'c0', 'tail')

    def __init__(self, model: str, c1: CellArg, c0: CellArg, tail: int = 0):
        if model == 'group':
            vars_ = {'group': GroupVar((c1, c0))}
        elif '%' in model:
            # assume 'group' is between, others are within
            raise NotImplementedError(f"model={model!r}: model with % for {self.__class__.__name__}")
        else:
            vars_ = None
        tail = tail_arg(tail)
        desc = '%s %s %s' % (c1, TAIL_REPR[tail], c0)
        Test.__init__(self, desc, '', vars=vars_, depend_on=[model])
        self.between_model = model
        self.c1 = c1
        self.c0 = c0
        self.tail = tail

    def as_dict(self):
        return {**Test.as_dict(self), 'model': self.between_model}

    def make(self, y, ds, force_permutation, kwargs):
        return testnd.TTestIndependent(y, self.between_model, self.c1, self.c0, 'subject', ds=ds, tail=self.tail, force_permutation=force_permutation, **kwargs)

    def make_uv(self, y, ds):
        return test.TTestIndependent(y, self.between_model, self.c1, self.c0, 'subject', ds=ds, tail=self.tail)


class TTestRelated(Test):
    """Related measures t-test

    Parameters
    ----------
    model : str
        The model which defines the cells that are used in the test. It is
        specified in the ``"x % y"`` format (like interaction definitions) where
        ``x`` and ``y`` are variables in the experiment's events.
    c1 : str | tuple
        The experimental condition. If the ``model`` is a single factor the
        condition is a :class:`str` specifying a value on that factor. If
        ``model`` is composed of several factors the cell is defined as a
        :class:`tuple` of :class:`str`, one value on each of the factors.
    c0 : str | tuple
        The control condition, defined like ``c1``.
    tail : int
        Tailedness of the test. ``0`` for two-tailed (default), ``1`` for upper tail
        and ``-1`` for lower tail.

    See Also
    --------
    MneExperiment.tests

    Examples
    --------
    Sample test definitions::

        tests = {
            'surprising=expected': TTestRelated('surprise', 'surprising', 'expected'),
        }

    Notes
    -----
    For a t-test between two epochs, use an
    :class:`~eelbrain.pipeline.EpochCollection` epoch and ``model='epoch'``.
    """
    kind = 'ttest_rel'
    DICT_ATTRS = Test.DICT_ATTRS + ('c1', 'c0', 'tail')

    def __init__(self, model: str, c1: CellArg, c0: CellArg, tail: int = 0):
        tail = tail_arg(tail)
        desc = '%s %s %s' % (c1, TAIL_REPR[tail], c0)
        Test.__init__(self, desc, model, cat=(c1, c0))
        self.c1 = c1
        self.c0 = c0
        self.tail = tail

    def make(self, y, ds, force_permutation, kwargs):
        return testnd.TTestRelated(y, self.model, self.c1, self.c0, 'subject', ds=ds, tail=self.tail, force_permutation=force_permutation, **kwargs)

    def make_uv(self, y, ds):
        return test.TTestRelated(y, self.model, self.c1, self.c0, 'subject', ds=ds, tail=self.tail)


class TContrastRelated(Test):
    """Contrasts of T-maps (see :class:`eelbrain.testnd.TContrastRelated`)

    Parameters
    ----------
    model : str
        The model which defines the cells that are used in the test. It is
        specified in the ``"x % y"`` format (like interaction definitions) where
        ``x`` and ``y`` are variables in the experiment's events.
    contrast : str
        Contrast specification using cells form the specified model (see
        :class:`eelbrain.testnd.TContrastRelated`)).
    tail : int
        Tailedness of the test. ``0`` for two-tailed (default), ``1`` for upper tail
        and ``-1`` for lower tail.

    See Also
    --------
    MneExperiment.tests

    Examples
    --------
    Sample test definitions::

        tests = {
            'a_b_intersection': TContrastRelated{'abc', 'min(a > c, b > c)', tail=1),
        }

    """
    kind = 't_contrast_rel'
    DICT_ATTRS = Test.DICT_ATTRS + ('contrast', 'tail')

    def __init__(self, model: str, contrast: str, tail: int = 0):
        tail = tail_arg(tail)
        Test.__init__(self, contrast, model)
        self.contrast = contrast
        self.tail = tail

    def make(self, y, ds, force_permutation, kwargs):
        return testnd.TContrastRelated(y, self.model, self.contrast, 'subject', ds=ds, tail=self.tail, force_permutation=force_permutation, **kwargs)


class ANOVA(Test):
    """ANOVA test

    Parameters
    ----------
    x : str
        ANOVA model specification, including ``subject`` for participant random
        effect (e.g., ``"x * y * subject"``; see :class:`eelbrain.test.ANOVA`).
    model : str
        Model for grouping trials before averaging (by default all fixed effects
        in ``x``). Should be specified in the ``"x % y"`` format (like
        interaction definitions) where ``x`` and ``y`` are variables in the
        experiment's events.
    vars : tuple | dict
        Variables to add dynamically.

    See Also
    --------
    MneExperiment.tests

    Examples
    --------
    Sample test definitions::

        tests = {
            'one_way': ANOVA('word_type * subject'),
            'two_way': ANOVA('word_type * meaning * subject'),
        }

    """
    kind = 'anova'
    DICT_ATTRS = Test.DICT_ATTRS + ('x',)

    def __init__(self, x: str, model: str = None, vars: dict = None):
        x_items = [item.strip() for item in x.split('*')]
        items = sorted(x_items)
        nested_in = (re.match(r'^subject\((\w+)\)?$', item) for item in items)
        between_items = []
        for match in filter(None, nested_in):
            between_item = match.group(1)
            items.remove(match.string)
            items.remove(between_item)
            between_items.append(between_item)
        if model is None:
            if 'subject' in items:
                items.remove('subject')
            elif not between_items:
                raise DefinitionError(f"x={x!r} without model: for mixed ANOVA, 'subject' needs to be in x; for between-subject ANOVA, model needs to be set explicitly")
            model = '%'.join(items)
        else:
            model_items = list(filter(None, (item.strip() for item in model.split('%'))))
            between_items.extend(set(items).difference(model_items))
        desc = ' * '.join(x_items)
        Test.__init__(self, desc, model, vars=vars, depend_on=between_items)
        self.x = '*'.join(x_items)

    def make(self, y, ds, force_permutation, kwargs):
        return testnd.ANOVA(y, self.x, ds=ds, force_permutation=force_permutation, **kwargs)

    def make_uv(self, y, ds):
        return test.ANOVA(y, self.x, ds=ds)


class TwoStageTest(Test):
    """Two-stage test: T-test of regression coefficients

    Stage 1: fit a regression model to the data for each subject.
    Stage 2: test coefficients from stage 1 against 0 across subjects.

    Parameters
    ----------
    stage_1 : str
        Stage 1 model specification. Coding for categorial predictors uses 0/1 dummy
        coding.
    vars : dict
        Add new variables for the stage 1 model. This is useful for specifying
        coding schemes based on categorial variables.
        Each entry specifies a variable with the following schema:
        ``{name: definition}``. ``definition`` can be either a string that is
        evaluated in the events-:class:`Dataset`, or a
        ``(source_name, {value: code})``-tuple (see example below).
        ``source_name`` can also be an interaction, in which case cells are joined
        with spaces (``"f1_cell f2_cell"``).
    model : str
        This parameter can be supplied to perform stage 1 tests on condition
        averages. If ``model`` is not specified, the stage1 model is fit on single
        trial data.

    See Also
    --------
    MneExperiment.tests

    Examples
    --------
    The first example assumes 2 categorical variables present in events,
    'a' with values 'a1' and 'a2', and 'b' with values 'b1' and 'b2'. These are
    recoded into 0/1 codes::

        TwoStageTest("a_num + b_num + a_num * b_num + index + a_num * index"},
                     vars={'a_num': ('a', {'a1': 0, 'a2': 1}),
                           'b_num': ('b', {'b1': 0, 'b2': 1})})

    The second test definition uses the "index" variable which is always present
    and specifies the chronological index of the events as an integer count.
    This variable can thus be used to test for a linear change over time. Due
    to the numeric nature of these variables interactions can be computed by
    multiplication::

        TwoStageTest("a_num + index + a_num * index",
                     vars={'a_num': ('a', {'a1': 0, 'a2': 1})

    Numerical variables can also defined using data-object methods (e.g.
    :meth:`Factor.label_length`) or from interactions::

        TwoStageTest('wordlength', vars={'wordlength': 'word.label_length()'})
        TwoStageTest("ab", vars={'ab': ('a%b', {'a1 b1': 0, 'a1 b2': 1, 'a2 b1': 1, 'a2 b2': 2})})
    """
    kind = 'two-stage'
    DICT_ATTRS = Test.DICT_ATTRS + ('stage_1',)

    def __init__(self, stage_1: str, vars: dict = None, model: str = None):
        Test.__init__(self, stage_1, model, vars=vars, depend_on=find_variables(stage_1))
        self.stage_1 = stage_1

    def make_stage_1(self, y, ds, subject, sub=None):
        """Assumes that model has already been applied"""
        return testnd.LM(y, self.stage_1, ds, subject=subject, sub=sub)

    @staticmethod
    def make_stage_2(lms, kwargs):
        lm = testnd.LMGroup(lms)
        lm.compute_column_ttests(**kwargs)
        return lm

    def make(self, y, ds, force_permutation, kwargs):
        lms = [self.make_stage_1(y, ds, subject, f"subject=={subject!r}") for subject in ds['subject'].cells]
        return self.make_stage_2(lms, kwargs)


TEST_CLASSES = {
    'anova': ANOVA,
    'ttest_1samp': TTestOneSample,
    'ttest_rel': TTestRelated,
    'ttest_ind': TTestIndependent,
    't_contrast_rel': TContrastRelated,
    'two-stage': TwoStageTest,
}


class TestDims:
    """Data shape for test

    Paremeters
    ----------
    string : str
        String describing data.
    time : bool
        Whether the base data contains a time axis.
    morph : bool
        If loading source space data, whether the data is morphed to the common
        brain.
    """
    RE = re.compile(r"^(source|sensor|meg|eeg)(?:\.(mean|rms))?$")
    source = None
    sensor = None

    def __init__(self, string, time=True, morph=False):
        self.time = bool(time)
        self.morph = bool(morph)
        m = self.RE.match(string)
        if m is None:
            raise ValueError(f"data={string!r}: invalid test dimension description")
        dim, aggregate = m.groups()
        if dim == 'meg':
            self._to_ndvar = ('mag',)
            self.y_name = 'meg'
            dim = 'sensor'
        elif dim == 'eeg':
            self._to_ndvar = ('eeg',)
            self.y_name = 'eeg'
            dim = 'sensor'
        elif dim == 'sensor':
            self._to_ndvar = None
            self.y_name = 'meg'
        elif dim == 'source':
            self._to_ndvar = None
            self.y_name = 'srcm' if self.morph else 'src'
        else:
            raise RuntimeError(f"dim={dim!r}")
        setattr(self, dim, aggregate or True)
        if sum(map(bool, (self.source, self.sensor))) != 1:
            raise ValueError(f"data={string!r}: invalid test dimension description")
        self.string = string

        dims = []
        if self.source is True:
            dims.append('source')
        elif self.sensor is True:
            dims.append('sensor')
        if self.time is True:
            dims.append('time')
        self.dims = tuple(dims)

        # whether parc is used from subjects or from common-brain
        if self.source is True:
            self.parc_level = 'common'
        elif self.source:
            self.parc_level = 'individual'
        else:
            self.parc_level = None

    @classmethod
    def coerce(cls, obj, time=True, morph=False):
        if isinstance(obj, cls):
            if obj.time == time and obj.morph == morph:
                return obj
            else:
                return cls(obj.string, time, morph)
        else:
            return cls(obj, time, morph)

    def __repr__(self):
        return "TestDims(%r)" % (self.string,)

    def __eq__(self, other):
        if not isinstance(other, TestDims):
            return False
        return self.string == other.string and self.time == other.time

    def data_to_ndvar(self, info):
        assert self.sensor
        if self._to_ndvar is None:
            return find_mne_channel_types(info)
        else:
            return self._to_ndvar


class ROITestResult:
    """Test results for temporal tests in one or more ROIs

    Attributes
    ----------
    subjects : tuple of str
        Subjects included in the test.
    samples : int
        ``samples`` parameter used for permutation tests.
    res : {str: NDTest} dict
        Test result for each ROI.
    n_trials_ds : Dataset
        Dataset describing how many trials were used in each condition per
        subject.
    """

    def __init__(self, subjects, samples, n_trials_ds, merged_dist, res):
        self.subjects = subjects
        self.samples = samples
        self.n_trials_ds = n_trials_ds
        self.merged_dist = merged_dist
        self.res = res

    def __getstate__(self):
        return {attr: getattr(self, attr) for attr in getfullargspec(self.__init__).args[1:]}

    def __setstate__(self, state):
        self.__init__(**state)


class ROI2StageResult(ROITestResult):
    """Test results for 2-stage tests in one or more ROIs

    Attributes
    ----------
    subjects : tuple of str
        Subjects included in the test.
    samples : int
        ``samples`` parameter used for permutation tests.
    res : {str: LMGroup} dict
        Test result for each ROI.
    n_trials_ds : Dataset
        Dataset describing how many trials were used in each condition per
        subject.
    """



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"Internet communication utilities"
from distutils.version import LooseVersion
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
import keyring
from keyring.errors import KeyringError
import smtplib
import socket
import sys
import traceback
import xmlrpc.client

from .system import user_activity
from .. import fmtxt
from . import ui


NOOB_DOMAIN = "Eelbrain"
NOOB_ADDRESS = 'n00b.eelbrain@gmail.com'


class KeychainError(Exception):
    "Error retrieving password from Keychain"


def check_for_update():
    """Check whether a new version of Eelbrain is available

    Prints a message if an update is available on the Python package index, does
    nothing otherwise.
    """
    current = sys.modules['eelbrain'].__version__
    if current == 'dev':
        print("Using Eelbrain development version")
        return
    pypi = xmlrpc.client.ServerProxy('https://pypi.python.org/pypi')
    versions = pypi.package_releases('eelbrain')
    newest = versions[-1]
    if LooseVersion(newest) > LooseVersion(current):
        print("New Eelbrain version available: %s (currently installed is %s)" %
              (newest, current))


def get_smtpserver(password, new_password=False):
    smtpserver = smtplib.SMTP_SSL('smtp.gmail.com', 465)
    smtpserver.ehlo()
    while True:
        try:
            smtpserver.login(NOOB_ADDRESS, password)
            if new_password:
                keyring.set_password(NOOB_DOMAIN, NOOB_ADDRESS, password)
            return smtpserver
        except smtplib.SMTPAuthenticationError:
            password = ui.ask_str("Eelbrain notifier password invalid. Please enter valid password.", "Notifier Password")
            if password:
                new_password = True
            else:
                raise


class Notifier:
    """
    A notification email sender supporting ``with`` statements

    Parameters
    ----------
    to : str
        Email address of the recipient.
    name : str
        Name of the job (will be included in subject line).
    crash_info_func : None | callable
        Will be called upon crash to produce a string that will be included
        in the crash report.
    debug : bool
        If the task crashes, start pdb instead of exiting.

    Examples
    --------
    To receive a message after a task has been executed:

    >>> notifier = Notifier('me@somewhere.com')
    >>> with notifier:
    ...     do_task()
    ...

    """
    def __init__(self, to, name='job', crash_info_func=None, debug=True):
        # get the password
        try:
            password = keyring.get_password(NOOB_DOMAIN, NOOB_ADDRESS)
        except KeyringError as error:
            raise KeychainError(f"""{error}

Notifier password could not be retrieved from Keychain. Try the following:

 - Open the Keychain application
 - Search for an item with the name 'Eelbrain'
 - Open the information on this item (select it and press command-i, or use the
   menu command 'File'/'Get Info')
 - Select the 'Access Control' panel
 - Select 'Allow all applications to access this item'
 - Save changes, exit, and try again
 
If this does not solve the issue, delete the item and repeat.
""")
        if password is None:
            password = ui.ask_str("Please enter the Eelbrain notifier password.", "Notifier Password")
            # test it
            print("Validating password...")
            smtpserver = get_smtpserver(password, True)
            smtpserver.close()

        self.to = to
        self.name = name
        self.crash_info_func = crash_info_func
        self.debug = debug
        self._password = password

    def __enter__(self):
        print("Notification enabled for %s..." % self.name)
        user_activity.__enter__()
        return self

    def __exit__(self, type_, value, traceback_):
        host = socket.gethostname()
        user_activity.__exit__(type_, value, traceback_)
        if isinstance(value, Exception):
            error = type_.__name__
            temp = '{host} encountered {error}: {value} in {task}'
            event = temp.format(host=host, error=error, value=value,
                                task=self.name)
            info = []

            # traceback
            tb_items = traceback.format_tb(traceback_)
            error_message = "%s: %s\n" % (error, value)
            tb_items.append(error_message)
            tb_str = '\n'.join(tb_items)
            info.append(tb_str)

            # object info
            if self.crash_info_func:
                info.extend(self.crash_info_func())

            self.send(event, info)

            # drop into pdb
            if self.debug:
                traceback.print_exc()
                print('')
                try:
                    import ipdb as pdb
                except ImportError:
                    import pdb
                pdb.post_mortem(traceback_)
        else:
            event = '{host} finished {task}'.format(host=host, task=self.name)
            self.send(event)

    def send(self, subject, info=()):
        """Send an email message

        Parameters
        ----------
        subject : str
            Email subject line.
        info : str | list of str | FMText
            Email body; if a list, successive entries are joined with three
            line breaks.
        """
        if isinstance(info, fmtxt.FMTextElement):
            msg = MIMEMultipart('alternative')
            # multipart message - the last part is preferred
            part1 = MIMEText(str(info), 'plain')
            part2 = MIMEText(fmtxt.html(info), 'html')
            msg.attach(part1)
            msg.attach(part2)
        elif isinstance(info, str):
            msg = MIMEText(info)
        else:
            msg = MIMEText('\n\n'.join(map(str, info)))

        host = socket.gethostname()
        msg['Subject'] = subject
        msg['From'] = f'Eelbrain on {host} <{NOOB_ADDRESS}>'
        msg['To'] = self.to

        try:
            smtpserver = get_smtpserver(self._password)
            smtpserver.sendmail(NOOB_ADDRESS, self.to, msg.as_string())
            smtpserver.close()
        except Exception as error:
            print(f"Could not send email because an error occurred, skipping notification\n\n{error}s")


class NotNotifier:
    # Helper to raise proper error message when user has not set owner attribute
    def __enter__(self):
        raise AttributeError("The notifier is disabled because the .owner attribute was not set")

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



"""Embedd Mayavi in Eelbrain

for testing:

src = datasets.get_mne_sample(src='ico', sub=[0])['src']
brain = plot.brain.brain(src.source, mask=False,hemi='lh',views='lat')
"""
from logging import getLogger

from mayavi.core.ui.api import SceneEditor, MlabSceneModel
import numpy as np
from traits.api import HasTraits, Instance
from traitsui.api import View, Item, HGroup, VGroup
from tvtk.api import tvtk
from tvtk.pyface.toolkit import toolkit_object

from .._wxgui import wx, ID, Icon
from .._wxgui.app import get_app
from .._wxgui.frame import EelbrainFrame
from .._wxgui.mpl_canvas import AxisLimitsDialog, SetTimeDialog

SCENE_NAME = 'scene_%i'
SURFACES = ('inflated', 'pial', 'smoothwm')

# undecorated scene
Scene = toolkit_object('scene:Scene')


class MayaviView(HasTraits):

    view = Instance(View)

    def __init__(self, width, height, n_rows, n_columns):
        HasTraits.__init__(self)

        n_scenes = n_rows * n_columns
        if n_scenes < 1:
            raise ValueError("n_rows=%r, n_columns=%r" % (n_rows, n_columns))

        self.scenes = tuple(MlabSceneModel() for _ in range(n_scenes))
        for i, scene in enumerate(self.scenes):
            self.add_trait(SCENE_NAME % i, scene)

        if n_rows == n_columns == 1:
            self.view = View(Item(SCENE_NAME % 0,
                                  editor=SceneEditor(scene_class=Scene),
                                  resizable=True, show_label=False),
                             width=width, height=height, resizable=True)
        else:
            rows = []
            for row in range(n_rows):
                columns = []
                for column in range(n_columns):
                    i = row * n_columns + column
                    item = Item(SCENE_NAME % i,
                                editor=SceneEditor(scene_class=Scene),
                                resizable=True, show_label=False)
                    columns.append(item)
                rows.append(HGroup(*columns))
            self.view = View(VGroup(*rows))

        self.figures = [scene.mayavi_scene for scene in self.scenes]


class BrainFrame(EelbrainFrame):
    _allow_user_set_title = True

    def __init__(self, parent, brain, title, width, height, n_rows, n_columns,
                 surf, pos):
        EelbrainFrame.__init__(self, parent, wx.ID_ANY, "Brain: %s" % title,
                               wx.DefaultPosition if pos is None else pos)

        # toolbar
        tb = self.CreateToolBar(wx.TB_HORIZONTAL)
        tb.SetToolBitmapSize(size=(32, 32))
        tb.AddTool(wx.ID_SAVE, "Save", Icon("tango/actions/document-save"))
        self.Bind(wx.EVT_TOOL, self.OnSaveAs, id=wx.ID_SAVE)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUISave, id=wx.ID_SAVE)
        # color-bar
        tb.AddTool(ID.PLOT_COLORBAR, "Plot Colorbar", Icon("plot/colorbar"))
        tb.Bind(wx.EVT_TOOL, self.OnPlotColorBar, id=ID.PLOT_COLORBAR)
        # surface
        self._surf_selector = wx.Choice(
            tb, choices=[name.capitalize() for name in SURFACES],
            name='Surface')
        if surf in SURFACES:
            self._surf_selector.SetSelection(SURFACES.index(surf))
        tb.AddControl(self._surf_selector, "Surface")
        self._surf_selector.Bind(
            wx.EVT_CHOICE, self.OnChoiceSurface, source=self._surf_selector)
        # view
        tb.AddTool(ID.VIEW_LATERAL, "Lateral View", Icon('brain/lateral'))
        self.Bind(wx.EVT_TOOL, self.OnSetView, id=ID.VIEW_LATERAL)
        tb.AddTool(ID.VIEW_MEDIAL, "Medial View", Icon('brain/medial'))
        self.Bind(wx.EVT_TOOL, self.OnSetView, id=ID.VIEW_MEDIAL)
        tb.AddTool(ID.SMOOTHING, "Smoothing Steps", Icon('brain/smoothing'))
        self.Bind(wx.EVT_TOOL, self.OnSetSmoothing, id=ID.SMOOTHING)
        # attach
        tb.AddStretchableSpace()
        tb.AddTool(ID.ATTACH, "Attach", Icon("actions/attach"))
        self.Bind(wx.EVT_TOOL, self.OnAttach, id=ID.ATTACH)
        tb.Realize()

        self.mayavi_view = MayaviView(width, height, n_rows, n_columns)
        self._n_rows = n_rows
        self._n_columns = n_columns
        # Use traits to create a panel, and use it as the content of this
        # wx frame.
        self.ui = self.mayavi_view.edit_traits(parent=self,
                                               view=self.mayavi_view.view,
                                               kind='subpanel')
        self.panel = self.ui.control
        # Hide the toolbar (the edit_traits command assigns scene_editor)
        for scene in self.mayavi_view.scenes:
            scene.interactor.interactor_style = tvtk.InteractorStyleTerrain()

        self.SetImageSize(width, height)

        self.figure = self.mayavi_view.figures
        self._brain = brain
        self.Bind(wx.EVT_CLOSE, self.OnClose)  # remove circular reference

        # replace key bindings
        self.panel.Unbind(wx.EVT_KEY_DOWN)
        for child in self.panel.Children[0].Children:
            panel = child.Children[0]
            panel.Unbind(wx.EVT_CHAR)
            panel.Bind(wx.EVT_KEY_DOWN, self.OnKeyDown)

    def CanCopy(self):
        return True

    def Copy(self):
        ss = self._brain.screenshot('rgba', True)
        ss = np.round(ss * 255).astype(np.uint8)
        h, w, _ = ss.shape
        image = wx.ImageFromDataWithAlpha(
            w, h, ss[:,:,:3].tostring(), ss[:,:,3].tostring())
        bitmap = image.ConvertToBitmap()
        data = wx.BitmapDataObject(bitmap)
        if not wx.TheClipboard.Open():
            getLogger('eelbrain').debug("Failed to open clipboard")
            return
        try:
            wx.TheClipboard.SetData(data)
        finally:
            wx.TheClipboard.Close()
            wx.TheClipboard.Flush()

    def OnAttach(self, event):
        get_app().Attach(self._brain, "Brain plot", 'brain', self)

    def OnChoiceSurface(self, event):
        self._brain._set_surf(SURFACES[event.GetSelection()])

    def OnClose(self, event):
        event.Skip()
        if self._brain is not None:
            self._brain._surfer_close()
            # remove circular references
            self._brain._frame = None
            self._brain = None

    def OnKeyDown(self, event):
        if self._brain is None:
            return  # plot is closed
        key = chr(event.GetUnicodeKey())
        if key == '.':
            self._brain._nudge_time(1)
        elif key == ',':
            self._brain._nudge_time(-1)
        else:
            event.Skip()

    def OnPlotColorBar(self, event):
        if self._brain._has_data():
            self._brain.plot_colorbar()
        elif self._brain._has_annot() or self._brain._has_labels():
            self._brain.plot_legend()

    def OnSave(self, event):
        self.OnSaveAs(event)

    def OnSaveAs(self, event):
        default_file = '%s.png' % self.GetTitle().replace(': ', ' - ')
        dlg = wx.FileDialog(self, "If no file type is selected below, it is "
                                  "inferred from the extension.",
                            defaultFile=default_file,
                            wildcard="Any (*.*)|*.*|PNG (*.png)|*.png",
                            style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
        if dlg.ShowModal() == wx.ID_OK:
            # no antialiasing (leads to loss of alpha channel)
            self._brain.save_image(dlg.GetPath(), 'rgba')
        dlg.Destroy()

    def OnSetSmoothing(self, event):
        props = self._brain.get_data_properties()
        old_value = props['smoothing_steps']
        dlg = wx.TextEntryDialog(self, "Data overlay smoothing steps:",
                                 "Smoothing Steps", str(old_value))
        value = None
        while True:
            if dlg.ShowModal() != wx.ID_OK:
                break
            try:
                value = int(dlg.GetValue())
                if value < 1:
                    raise ValueError("Needs to be at least 1")
                elif value > 100:
                    raise ValueError("Values > 100 take too long")
            except Exception as exception:
                msg = wx.MessageDialog(self, str(exception), "Invalid Entry",
                                       wx.OK | wx.ICON_ERROR)
                msg.ShowModal()
                msg.Destroy()
            else:
                break
        dlg.Destroy()
        if value is not None and value != old_value:
            self._brain.set_data_smoothing_steps(value)

    def OnSetView(self, event):
        if event.Id == ID.VIEW_LATERAL:
            views = ('lateral', 'medial')
        elif event.Id == ID.VIEW_MEDIAL:
            views = ('medial', 'lateral')
        else:
            return

        for row, view in zip(self._brain.brain_matrix, views):
            for b in row:
                b.show_view(view)

    def OnSetVLim(self, event):
        vlim = self._brain.get_vlim()
        dlg = AxisLimitsDialog(vlim, None, None, self)
        if dlg.ShowModal() == wx.ID_OK:
            self._brain.set_vlim(*dlg.vlim)
        dlg.Destroy()

    def OnSetTime(self, event):
        current_time = self._brain.get_time()
        dlg = SetTimeDialog(self, current_time)
        if dlg.ShowModal() == wx.ID_OK:
            self._brain.set_time(dlg.time)
        dlg.Destroy()

    def OnUpdateUISave(self, event):
        event.Enable(True)

    def OnUpdateUISaveAs(self, event):
        event.Enable(True)

    def OnUpdateUISetVLim(self, event):
        event.Enable(self._brain._has_data())

    def OnUpdateUISetTime(self, event):
        event.Enable(self._brain._has_data())

    def SetImageSize(self, width, height):
        if self._n_columns == 1 and self._n_rows == 1:
            width += 2
            height += 2
        else:
            width += self._n_columns * 2 + 4
            height += self._n_rows * 2 + 4
        self.SetClientSize((width, height))



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Statistical tests for univariate variables"""
from functools import partial
import itertools
from typing import Literal, Sequence, Union

import numpy as np
import scipy.stats

from .. import fmtxt
from .._celltable import Celltable
from .._data_obj import (
    CategorialArg, CellArg, IndexArg, VarArg, NumericArg,
    Dataset, Factor, Interaction, Var, NDVar,
    ascategorial, asfactor, asnumeric, assub, asvar, asndvar,
    combine,
    cellname, dataobj_repr, nice_label,
)
from .._utils import LazyProperty
from .permutation import resample
from . import stats


__test__ = False
DEFAULT_LEVELS = {.05: '*', .01: '**', .001: '***'}
DEFAULT_LEVELS_TREND = {.05: '*', .01: '**', .001: '***', .1: '`'}

MCCArg = Union[bool, Literal['hochberg', 'bonferroni', 'holm']]


def get_levels(
        levels: Union[bool, dict],
        trend: Union[bool, str] = False,
):
    if levels is True:
        if trend is True:
            return DEFAULT_LEVELS_TREND
        elif trend:
            return {**DEFAULT_LEVELS, .1: trend}
        else:
            return DEFAULT_LEVELS
    elif trend:
        raise TypeError("trend=%r only valid when levels=True" % (trend,))
    return levels


class Correlation:
    """Pearson product moment correlation between y and x

    Parameters
    ----------
    y : Var | NDVar
        First variable.
    x : Var | NDVar
        Second variable. Needs to have same type/shape as ``y``.
    sub : index
        Use only a subset of the data
    ds : Dataset
        If a Dataset is given, all data-objects can be specified as names of
        Dataset variables.

    Attributes
    ----------
    r : float
        Pearson correlation coefficient.
    p : float
        Two-tailed p-value.
    df : int
        Degrees of freedom.
    """
    def __init__(
            self,
            y: NumericArg,
            x: NumericArg,
            sub: IndexArg = None,
            ds: Dataset = None,
    ):
        sub, n = assub(sub, ds, return_n=True)
        y, n = asnumeric(y, sub, ds, n, return_n=True, array=True)
        x = asnumeric(x, sub, ds, n, array=True)
        if type(y) is not type(x):
            raise TypeError(f"y and x must be same type; got type(y)={type(y)}, type(x)={type(x)}")
        elif isinstance(y, Var):
            x_y = y.x
            x_x = x.x
        elif isinstance(y, NDVar):
            if y.dims != x.dims:
                raise ValueError(f"y and x have different dimensions; y.dims={y.dims}, x.dims={x.dims}")
            x_y = y.x.ravel()
            x_x = x.x.ravel()
        else:
            if y.shape != x.shape:
                raise ValueError(f"Shape mismatch: y.shape={y.shape}, x.shape={x.shape}")
            x_y = y.ravel()
            x_x = x.ravel()
        self.r, self.p, self.df = _corr(x_y, x_x)
        self._y = dataobj_repr(y)
        self._x = dataobj_repr(x)

    def __repr__(self):
        return f"<Correlation: {self._y} ~ {self._x}; {self._asfmtext()}> "

    def _asfmtext(self, **_):
        return fmtxt.FMText([fmtxt.eq('r', self.r, self.df), ', ', fmtxt.peq(self.p)])


def lilliefors(data, formatted=False, **kwargs):
    """Lilliefors' test for normal distribution

    The Lilliefors test is an adaptation of the Kolmogorov-Smirnov test. It
    is used to test the null hypothesis that data come from a normally
    distributed population, when the null hypothesis does not specify which
    normal distribution, i.e. does not specify the expected value and variance.

    Parameters
    ----------
    data : array_like

    formatted : bool
        Return a single string with the results instead of the numbers.
    kwargs :
        All keyword arguments are forwarded to :func:`scipy.stats.kstest`.

    Returns
    -------
    D : float
        The D-value of the Kolmogorov-Smirnov Test
    p_estimate : float
        The approximate p value according to Dallal and Wilkinson (1986).
        Requires minimal sample size of 5. p is reasonably accurate only when
        it is <= .1 (cf. Dallal and Wilkens).

    Notes
    -----
    Uses the scipy.stats.kstest implementation of the Kolmogorov-Smirnov test.

    References
    ----------
    Dallal, G. E. and Wilkinson, L. (1986). An Analytic Approximation to the
            Distribution of Lilliefors's Test Statistic for Normality. The
            American Statistician, 40(4), 294--296.
    Lilliefors, H. W. (1967). On the Kolmogorov-Smirnov Test for Normality
            with Mean and Variance Unknown. Journal of the American
            Statistical Association, 62(318), 399--402.
    """
    # p values agree with R lillie.test (nortest package) on low values of p.
    # lillie.test adjusts something at p>.1
    # http://pbil.univ-lyon1.fr/library/nortest/R/nortest
    data = np.asarray(data)
    N = len(data)  # data.shape[-1] #axis]
    assert N >= 5, "sample size must be greater than 4"
    # perform Kolmogorov-Smirnov with estimated mean and std
    m = np.mean(data)  # , axis=axis)
    s = np.std(data, ddof=1)  # , axis=axis)
    D, ks_p = scipy.stats.kstest(data, 'norm', args=(m, s), **kwargs)
    # approximate p (Dallal)
    if N > 100:
        D *= (N / 100) ** .49
        N = 100
    p_estimate = np.exp(- 7.01256 * D ** 2 * (N + 2.78019)
                        + 2.99587 * D * (N + 2.78019) ** .5
                        - .122119
                        + .974598 / (N ** .5)
                        + 1.67997 / N)
    # approximate P (Molin & Abdi)
    L = D  # ???
    b2 = 0.08861783849346
    b1 = 1.30748185078790
    b0 = 0.37872256037043
    A = (-(b1 + N) + np.sqrt((b1 + N) ** 2 - 4 * b2 * (b0 - L ** -2))) / 2 * b2
    Pr = (-.37782822932809
          + 1.67819837908004 * A
          - 3.02959249450445 * A ** 2
          + 2.80015798142101 * A ** 3
          - 1.39874347510845 * A ** 4
          + 0.40466213484419 * A ** 5
          - 0.06353440854207 * A ** 6
          + 0.00287462087623 * A ** 7
          + 0.00069650013110 * A ** 8
          - 0.00011872227037 * A ** 9
          + 0.00000575586834 * A ** 10)
    if formatted:
        txt = "D={0:.4f}, Dallal p={1:.4f}, Molin&Abdi p={2:.4f}"
        return txt.format(D, p_estimate, Pr)
    else:
        return D, p_estimate


def _hochberg_threshold(N, alpha=.05):
    j = np.arange(N)
    threshold = alpha / (N - j)
    return threshold


def mcp_adjust(ps, method='Hochberg'):
    """Adjust p-values for multiple comparison

    Parameters
    ----------
    ps : sequence of scalar
        P-values.
    method : 'hochberg' | 'bonferroni' | 'holm'
        Correction method. Default is 'hochberg'.

    Returns
    -------
    adjusted_ps : list of scalar
        Adjusted p-values.
    """
    if not method:
        return ps
    n = len(ps)
    if n <= 1:
        return ps
    method_ = method.lower()
    if method_ == 'bonferroni':
        ps_adjusted = [p * n for p in ps]
    elif method_ in ('hochberg', 'holm'):
        ascsort = np.argsort(ps)
        ps_asc = np.array(ps)[ascsort]
        iout_asc = np.arange(n)[ascsort]
        ps_adjusted = [-1] * n
        p_buffer = 1
        if method_ == 'holm':
            for i in range(n):
                p = ps_asc[i]
                p_adj = (n - i) * p
                p_buffer = max(p_buffer, p_adj)
                ps_adjusted[iout_asc[i]] = p_buffer
        elif method_ == 'hochberg':
            for i in range(1, n + 1):
                p = ps_asc[-i]
                p_adj = (i) * p
                p_buffer = min(p_adj, p_buffer)
                ps_adjusted[iout_asc[-i]] = p_buffer
    else:
        msg = ('%r is not a valid argument for multiple comparison correction '
               'method' % method)
        raise ValueError(msg)

    return ps_adjusted


def _get_correction_caption(corr, n):
    if corr == 'Hochberg':
        return "(* Corrected after Hochberg, 1988)"
    elif corr == 'Bonferroni':
        return "(* Bonferroni corrected for %i tests)" % n
    elif corr == 'Holm':
        return "(* Corrected after Holm)"
    else:
        return "(* Uncorrected)"


def _n_stars(p: float, levels: Sequence[float] = (0.05, 0.01, 0.001)):
    return sum(p <= level for level in levels)


def star(
        p_list: Union[float, Sequence[float]],
        out: type = str,
        levels: Union[bool, dict] = True,
        trend: Union[bool, str] = False,
        eq_strlen: bool = False,
):
    """Determine number of stars for p-value

    Parameters
    ----------
    p_list
        P-values.
    out : {str, int}
        Return string with stars ('**') or an integer indicating the number of
        stars.
    levels
        ``{p: str, ...}`` dictionary. The default is ``{.05 : '*',
        .01 : '**', .001: '***'}``; ``trend=True`` adds ``{.1: "'"}``.
    trend
        Add trend to default ``levels``.
    eq_strlen
        Equalize string lengths; when strings are returned, make sure they all
        have the same length (default ``False``).
    """
    levels = get_levels(levels, trend)

    levels_descending = sorted(levels.keys(), reverse=True)
    symbols_descending = [''] + [levels[l] for l in levels_descending]

    if out is int:
        int_out = True
    elif out is str:
        int_out = False
    else:
        raise TypeError("out=%r" % (out,))

    # allow input (p_list) to contain single p-value
    if not np.iterable(p_list):
        n = _n_stars(p_list, levels)
        if int_out:
            return out
        else:
            return symbols_descending[n]

    ns = [_n_stars(p, levels) for p in p_list]
    if int_out:
        return ns
    symbols = [symbols_descending[n] for n in ns]
    if eq_strlen:
        maxlen = max(map(len, symbols))
        return [s.ljust(maxlen) for s in symbols]
    else:
        return symbols


def star_factor(p, levels=True):
    """Create a factor with stars for a sequence of p-values

    Parameters
    ----------
    p : sequence of scalar
        Sequence of p-values.
    levels : dict {scalar: str}
        {value: star-mark} dictionary.

    Returns
    -------
    stars : Factor
        Factor with the appropriate star marking for each item in p.
    """
    levels = get_levels(levels)
    sorted_levels = sorted(levels, reverse=True)
    star_labels = {i: levels[v] for i, v in enumerate(sorted_levels, 1)}
    star_labels[0] = ''
    level_values = np.reshape(sorted_levels, (-1, 1))
    return Factor(np.sum(p <= level_values, 0), labels=star_labels)


def _independent_measures_args(y, x, c1, c0, match, ds, sub, nd_data=False):
    "Interpret parameters for independent measures tests (2 different argspecs)"
    if isinstance(x, str):
        x = ds.eval(x)

    if nd_data:
        two_y = isinstance(x, NDVar)
        coerce = asndvar
    else:
        two_y = isinstance(x, Var) and match is None
        coerce = asvar

    if two_y:
        assert match is None
        y1 = coerce(y, sub, ds)
        y0 = coerce(x, sub, ds)
        y = combine((y1, y0))
        c1_name = y1.name if c1 is None else c1
        c0_name = y0.name if c0 is None else c0
        x_name = y0.name
    else:
        ct = Celltable(y, x, match, sub, cat=(c1, c0), ds=ds, coercion=coerce, dtype=np.float64)
        c1, c0 = ct.cat
        c1_name = c1
        c0_name = c0
        x_name = ct.x.name
        match = ct.match
        y = ct.y
        y1 = ct.data[c1]
        y0 = ct.data[c0]
    return y, y1, y0, c1, c0, match, x_name, c1_name, c0_name


def _related_measures_args(y, x, c1, c0, match, ds, sub, nd_data=False):
    "Interpret parameters for related measures tests (2 different argspecs)"
    if isinstance(x, str):
        if ds is None:
            raise TypeError(f"x={x!r} specified as str without specifying ds")
        x = ds.eval(x)

    if nd_data:
        two_y = isinstance(x, NDVar)
        coerce = asndvar
    else:
        two_y = isinstance(x, Var) and match is None
        coerce = asvar

    if two_y:
        assert match is None
        y1 = coerce(y, sub, ds)
        n = len(y1)
        y0 = coerce(x, sub, ds, n)
        c1_name = y1.name if c1 is None else c1
        c0_name = y0.name if c0 is None else c0
        x_name = y0.name
    elif match is None:
        raise TypeError("The `match` argument needs to be specified for related measures tests")
    else:
        ct = Celltable(y, x, match, sub, cat=(c1, c0), ds=ds, coercion=coerce, dtype=np.float64)
        c1, c0 = ct.cat
        c1_name = c1
        c0_name = c0
        if not ct.all_within:
            raise ValueError(f"conditions {c1!r} and {c0!r} do not have the same values on {dataobj_repr(ct.match)}")
        n = len(ct.y) // 2
        y1 = ct.y[:n]
        y0 = ct.y[n:]
        x_name = ct.x.name
        match = ct.match
    return y1, y0, c1, c0, match, n, x_name, c1_name, c0_name


def ttest(y, x=None, against=0, match=None, sub=None, corr='Hochberg',
          title='{desc}', ds=None):
    """T-tests for one or more samples

    parameters
    ----------
    y : Var
        Dependent variable
    x : None | categorial
        Perform tests separately for all categories in x.
    against : scalar | str | tuple
        Baseline against which to test (scalar or category in x).
    match : None | Factor
        Repeated measures factor.
    sub : index
        Only use part of the data.
    corr : None | 'hochberg' | 'bonferroni' | 'holm'
        Method for multiple comparison correction (default 'hochberg').
    title : str
        Title for the table.
    ds : Dataset
        If a Dataset is given, all data-objects can be specified as names of
        Dataset variables

    Returns
    -------
    table : FMText Table
        Table with results.
    """
    ct = Celltable(y, x, match, sub, ds=ds, coercion=asvar)

    par = True
    if par:
        infl = '' if x is None else 's'
        if ct.y.name is None:
            title_desc = "T-test%s against %s" % (infl, cellname(against))
        else:
            title_desc = "T-test%s of %s against %s" % (infl, ct.y.name, cellname(against))
        statistic_name = 't'
    else:
        raise NotImplementedError

    names = []
    diffs = []
    ts = []
    dfs = []
    ps = []

    if isinstance(against, (str, tuple)):
        if against not in ct.cells:
            x_repr = 'x' if ct.x.name is None else repr(ct.x.name)
            raise ValueError("agains=%r: %r is not a cell in %s"
                             % (against, against, x_repr))
        k = len(ct.cells) - 1
        baseline = ct.data[against]

        for cell in ct.cells:
            if cell == against:
                continue
            names.append(cell)
            if match is not None and ct.within[cell, against]:
                diffs.append(ct.data[cell].mean() - baseline.mean())
                t, p = scipy.stats.ttest_rel(ct.data[cell], baseline)
                df = len(baseline) - 1
            else:
                data = ct.data[cell]
                diffs.append(data.mean() - baseline.mean())
                t, p = scipy.stats.ttest_ind(data, baseline)
                df = len(baseline) + len(data) - 2
            ts.append(t)
            dfs.append(df)
            ps.append(p)

    elif np.isscalar(against):
        k = len(ct.cells)

        for cell in ct.cells:
            label = cellname(cell)
            data = ct.data[cell].x
            t, p = scipy.stats.ttest_1samp(data, against)
            df = len(data) - 1
            names.append(label)
            diffs.append(data.mean() - against)
            ts.append(t)
            dfs.append(df)
            ps.append(p)
    else:
        raise TypeError("against=%s" % repr(against))

    if k <= 1:
        corr = None

    if corr:
        ps_adjusted = mcp_adjust(ps, corr)
    else:
        ps_adjusted = ps
    stars = star(ps_adjusted, out=str)
    if len(set(dfs)) == 1:
        df_in_header = True
    else:
        df_in_header = False

    table = fmtxt.Table('l' + 'r' * (4 - df_in_header + bool(corr)))
    table.title(title.format(desc=title_desc))
    if corr:
        table.caption(_get_correction_caption(corr, k))

    # header
    table.cell("Effect")
    table.cell("Difference")
    if df_in_header:
        table.cell(fmtxt.symbol(statistic_name, dfs[0]))
    else:
        table.cell(statistic_name, 'math')
        table.cell('df', 'math')
    table.cell('p', 'math')
    if corr:
        table.cell(fmtxt.symbol('p', corr))
    table.midrule()

    # body
    for name, diff, t, mark, df, p, p_adj in zip(names, diffs, ts, stars, dfs, ps, ps_adjusted):
        table.cell(name)
        table.cell(diff)
        table.cell(fmtxt.stat(t, stars=mark, of=3))
        if not df_in_header:
            table.cell(df)

        table.cell(fmtxt.p(p))
        if corr:
            table.cell(fmtxt.p(p_adj))
    return table


class TTest:

    def __init__(self, difference, t, df, tail):
        self._difference = difference
        self.t = t
        self.df = df
        self.p = stats.ttest_p(self.t, self.df, tail)
        self.tail = tail

    def _asfmtext(
            self,
            rasterize: bool = None,
            difference: bool = False,
            **_,
    ):
        out = [fmtxt.eq('t', self.t, self.df), ', ', fmtxt.peq(self.p)]
        if difference:
            out = [fmtxt.eq('d', self._difference), ', '] + out
        return fmtxt.FMText(out)


class TTestOneSample(TTest):
    """One-sample t-test

    Parameters
    ----------
    y : Var
        Dependent variable.
    match : categorial
        Units within which measurements are related (e.g. 'subject' in a
        within-subject comparison).
    sub : index-array
        Perform the test with a subset of the data.
    ds : Dataset
        If a Dataset is specified, all data-objects can be specified as
        names of Dataset variables.
    popmean : float
        Population mean to test against (default 0).
    tail : 0 | 1 | -1
        Which tail of the t-distribution to consider:
        0: both (two-tailed, default);
        1: upper tail (one-tailed);
        -1: lower tail (one-tailed).

    Attributes
    ----------
    mean : float
        Mean of ``y``.
    t : float
        T-value.
    p : float
        P-value.
    tail : 0 | 1 | -1
        Tailedness of the p value.
    df : int
        Degrees of freedom.
    full : FMText
        Full description of the test result.
    """
    def __init__(
            self,
            y: VarArg,
            match: CategorialArg = None,
            sub: IndexArg = None,
            ds: Dataset = None,
            popmean: float = 0,
            tail: int = 0,
    ):
        ct = Celltable(y, None, match, sub, ds=ds, coercion=asvar)
        n = len(ct.y)
        if n <= 2:
            raise ValueError(f"Not enough observations for t-test ({n=})")

        self.mean = ct.y.mean()
        self.popmean = popmean
        self._y = ct.y
        self._y_name = dataobj_repr(ct.y)
        v = ct.y.x[:, None]
        if popmean:
            v = v - popmean
        t = stats.t_1samp(v)[0]
        TTest.__init__(self, v.mean(), t, n - 1, tail)

    def __repr__(self):
        cmp = '=><'[self.tail]
        return f"<{self.__class__.__name__}: {self._y_name} {cmp} {self.popmean}; {self._asfmtext(difference=True)}>"

    @LazyProperty
    def full(self):
        return fmtxt.FMText([fmtxt.eq('M', self.mean), ', ', fmtxt.eq('SD', self._y.std()), ', ', self._asfmtext()])


class TTestIndependent(TTest):
    """Independent measures t-test

    The test data can be specified in two forms:

     - In long form, with ``y`` supplying the data, ``x`` specifying condition
       for each case.
     - With ``y`` and ``x`` supplying data for the two conditions.

    Parameters
    ----------
    y : Var
        Dependent variable.
    x : categorial
        Model containing the cells which should be compared.
    c1 : str | tuple | None
        Test condition (cell of ``x``). ``c1`` and ``c0`` can be omitted if
        ``x`` only contains two cells, in which case cells will be used in
        alphabetical order.
    c0 : str | tuple | None
        Control condition (cell of ``x``).
    match : categorial
        Units within which measurements are related and should be averaged over
        (e.g. 'subject' in a between-group comparison).
    sub : index-array
        Perform the test with a subset of the data.
    ds : Dataset
        If a Dataset is specified, all data-objects can be specified as
        names of Dataset variables.
    tail : 0 | 1 | -1
        Which tail of the t-distribution to consider:
        0: both (two-tailed, default);
        1: upper tail (one-tailed);
        -1: lower tail (one-tailed).

    Attributes
    ----------
    t : float
        T-value.
    p : float
        P-value.
    tail : 0 | 1 | -1
        Tailedness of the p value.
    df : int
        Degrees of freedom.
    full : FMText
        Full description of the test result.
    """
    def __init__(
            self,
            y: VarArg,
            x: CategorialArg,
            c1: CellArg = None,
            c0: CellArg = None,
            match: CategorialArg = None,
            sub: IndexArg = None,
            ds: Dataset = None,
            tail: int = 0,
    ):
        y, y1, y0, c1, c0, match, x_name, c1_name, c0_name = _independent_measures_args(y, x, c1, c0, match, ds, sub)

        n1 = len(y1)
        n = len(y)
        df = n - 2
        groups = np.arange(n) < n1
        d = y[groups].mean() - y[~groups].mean()
        groups.dtype = np.int8
        t = stats.t_ind(y.x[:, None], groups)[0]

        self._y = dataobj_repr(y)
        self._x = x_name
        TTest.__init__(self, d, t, df, tail)
        self.c1_name = c1_name
        self.c0_name = c0_name
        self._y1 = y1
        self._y0 = y0
        self._two_y = c1 is None

    def __repr__(self):
        cmp = '=><'[self.tail]
        if self._two_y:
            desc = f"{self.c1_name} {cmp} {self.c0_name}"
        else:
            desc = f"{self._y} ~ {self._x}, {self.c1_name} {cmp} {self.c0_name}"
        return f"<{self.__class__.__name__}: {desc}; {self._asfmtext(difference=True)}>"

    @LazyProperty
    def full(self):
        return fmtxt.FMText([
            self.c1_name, ': ', fmtxt.eq('M', self._y1.mean()), ', ', fmtxt.eq('SD', self._y1.std()), '; ',
            self.c0_name, ': ', fmtxt.eq('M', self._y0.mean()), ', ', fmtxt.eq('SD', self._y0.std()), '; ',
            self._asfmtext()])


class MannWhitneyU:
    """Mann-Whitney U-test (non-parametric independent measures test)

    The test data can be specified in two forms:

     - In long form, with ``y`` supplying the data, ``x`` specifying condition
       for each case.
     - With ``y`` and ``x`` supplying data for the two conditions.

    Parameters
    ----------
    y : Var
        Dependent variable. Alternatively, the first of two variables that are
        compared.
    x : categorial
        Model containing the cells which should be compared. Alternatively, the
        second of two varaibles that are compared.
    c1 : str | tuple | None
        Test condition (cell of ``x``). ``c1`` and ``c0`` can be omitted if
        ``x`` only contains two cells, in which case cells will be used in
        alphabetical order.
    c0 : str | tuple | None
        Control condition (cell of ``x``).
    match : categorial
        Units within which measurements are related (e.g. 'subject' in a
        within-subject comparison). If match is unspecified, it is assumed that
        ``y`` and ``x`` are two measurements with matched cases.
    sub : index-array
        Perform the test with a subset of the data.
    ds : Dataset
        If a Dataset is specified, all data-objects can be specified as
        names of Dataset variables.
    tail : 0 | 1 | -1
        Which tail of the t-distribution to consider:
        0: both (two-tailed, default);
        1: upper tail (one-tailed);
        -1: lower tail (one-tailed).
    continuity : bool
        Continuity correction (default ``True``).

    Attributes
    ----------
    u : float
        Mann-Whitney U statistic.
    p : float
        P-value.
    tail : 0 | 1 | -1
        Tailedness of the p value.

    See Also
    --------
    TTestRelated : parametric alternative

    Notes
    -----
    Based on :func:`scipy.stats.mannwhitneyu`.
    """
    def __init__(
            self,
            y: VarArg,
            x: Union[CategorialArg, VarArg],
            c1: CellArg = None,
            c0: CellArg = None,
            match: CategorialArg = None,
            sub: IndexArg = None,
            ds: Dataset = None,
            tail: int = 0,
            continuity: bool = True,
    ):
        y, y1, y0, c1, c0, match, x_name, c1_name, c0_name = _independent_measures_args(y, x, c1, c0, match, ds, sub)
        if tail == 0:
            alternative = 'two-sided'
        elif tail == 1:
            alternative = 'greater'
        elif tail == -1:
            alternative = 'less'
        else:
            raise ValueError(f"tail={tail!r}")
        self.u, self.p = scipy.stats.mannwhitneyu(y1.x, y0.x, continuity, alternative)
        self._y = dataobj_repr(y1)
        self._x = x_name
        self._c1 = c1
        self._c0 = c0
        self.tail = tail
        self.continuity = continuity
        self._two_y = c1 is None

    def __repr__(self):
        cmp = '=><'[self.tail]
        if self._two_y:
            desc = f"{self._c1} {cmp} {self._c0}"
        else:
            desc = f"{self._y} ~ {self._x}, {self._c1} {cmp} {self._c0}"
        return f"<MannWhitneyU: {desc}; {self._asfmtext()}>"

    def _asfmtext(self, **_):
        return fmtxt.FMText([fmtxt.eq('U', self.u), ', ', fmtxt.peq(self.p)])


class TTestRelated(TTest):
    """Related-measures t-test

    The test data can be specified in two forms:

     - In long form, with ``y`` supplying the data, ``x`` specifying condition
       for each case and ``match`` determining which cases are related.
     - In wide/repeated measures form, with ``y`` and ``x`` both supplying data
       with matching case order.

    Parameters
    ----------
    y : Var
        Dependent variable. Alternatively, the first of two variables that are
        compared.
    x : categorial
        Model containing the cells which should be compared. Alternatively, the
        second of two varaibles that are compared.
    c1 : str | tuple | None
        Test condition (cell of ``x``). ``c1`` and ``c0`` can be omitted if
        ``x`` only contains two cells, in which case cells will be used in
        alphabetical order.
    c0 : str | tuple | None
        Control condition (cell of ``x``).
    match : categorial
        Units within which measurements are related (e.g. 'subject' in a
        within-subject comparison). If match is unspecified, it is assumed that
        ``y`` and ``x`` are two measurements with matched cases.
    sub : index-array
        Perform the test with a subset of the data.
    ds : Dataset
        If a Dataset is specified, all data-objects can be specified as
        names of Dataset variables.
    tail : 0 | 1 | -1
        Which tail of the t-distribution to consider:
        0: both (two-tailed, default);
        1: upper tail (one-tailed);
        -1: lower tail (one-tailed).

    Attributes
    ----------
    t : float
        T-value.
    p : float
        P-value.
    tail : 0 | 1 | -1
        Tailedness of the p value.
    difference : Var
        Difference values.
    df : int
        Degrees of freedom.
    c1_mean : float
        Mean of condition ``c1``.
    c0_mean : float
        Mean of condition ``c0``.
    full : FMText
        Full description of the test result.

    See Also
    --------
    WilcoxonSignedRank : non-parametric alternative
    """
    def __init__(
            self,
            y: VarArg,
            x: Union[CategorialArg, VarArg],
            c1: CellArg = None,
            c0: CellArg = None,
            match: CategorialArg = None,
            sub: IndexArg = None,
            ds: Dataset = None,
            tail: int = 0,
    ):
        y1, y0, c1, c0, match, n, x_name, c1_name, c0_name = _related_measures_args(y, x, c1, c0, match, ds, sub)
        if n <= 2:
            raise ValueError("Not enough observations for t-test (n=%i)" % n)
        self._y = dataobj_repr(y1)
        self._x = x_name
        self._c1 = c1
        self._c0 = c0
        self.c1_name = c1_name
        self.c0_name = c0_name
        self.c1_mean = y1.mean()
        self.c0_mean = y0.mean()
        self.difference = y1 - y0
        t = stats.t_1samp(self.difference.x[:, None])[0]
        TTest.__init__(self, self.difference.x.mean(), t, n - 1, tail)
        self._match = dataobj_repr(match, True)

    def __repr__(self):
        cmp = '=><'[self.tail]
        if self._match is None:
            desc = f"{self._c1} {cmp} {self._c0}"
        else:
            desc = f"{self._y} ~ {self._x}, {self._c1} {cmp} {self._c0}"
        return f"<{self.__class__.__name__}: {desc}; {self._asfmtext(difference=True)}>"

    @LazyProperty
    def full(self):
        return fmtxt.FMText([
            self.c1_name, ': ', fmtxt.eq('M', self.c1_mean), '; ',
            self.c0_name, ': ', fmtxt.eq('M', self.c0_mean), '; ',
            'difference: ', fmtxt.eq('M', self.difference.mean()), ', ',
            fmtxt.eq('SD', self.difference.std()), ', ',
            self._asfmtext()])


class WilcoxonSignedRank:
    """Wilcoxon signed-rank test (non-parametric related measures test)

    The test data can be specified in two forms:

     - In long form, with ``y`` supplying the data, ``x`` specifying condition
       for each case and ``match`` determining which cases are related.
     - In wide/repeated measures form, with ``y`` and ``x`` both supplying data
       with matching case order.

    Parameters
    ----------
    y : Var
        Dependent variable. Alternatively, the first of two variables that are
        compared.
    x : categorial
        Model containing the cells which should be compared. Alternatively, the
        second of two varaibles that are compared.
    c1 : str | tuple | None
        Test condition (cell of ``x``). ``c1`` and ``c0`` can be omitted if
        ``x`` only contains two cells, in which case cells will be used in
        alphabetical order.
    c0 : str | tuple | None
        Control condition (cell of ``x``).
    match : categorial
        Units within which measurements are related (e.g. 'subject' in a
        within-subject comparison). If match is unspecified, it is assumed that
        ``y`` and ``x`` are two measurements with matched cases.
    sub : index-array
        Perform the test with a subset of the data.
    ds : Dataset
        If a Dataset is specified, all data-objects can be specified as
        names of Dataset variables.
    tail : 0 | 1 | -1
        Which tail of the t-distribution to consider:
        0: both (two-tailed, default);
        1: upper tail (one-tailed);
        -1: lower tail (one-tailed).
    zero_method : str
        How to handle zero differences (see :func:`scipy.stats.wilcoxon`).
    correction : bool
        Continuity correction (default ``False``).

    Attributes
    ----------
    w : float
        Rank sum statistic.
    p : float
        P-value.
    tail : 0 | 1 | -1
        Tailedness of the p value.
    difference : Var
        Difference values.
    c1_mean : float
        Mean of condition ``c1``.
    c0_mean : float
        Mean of condition ``c0``.

    See Also
    --------
    TTestRelated : parametric alternative

    Notes
    -----
    Based on :func:`scipy.stats.wilcoxon`.
    """
    def __init__(
            self,
            y: VarArg,
            x: Union[CategorialArg, VarArg],
            c1: CellArg = None,
            c0: CellArg = None,
            match: CategorialArg = None,
            sub: IndexArg = None,
            ds: Dataset = None,
            tail: int = 0,
            zero_method: str = 'wilcox',
            correction: bool = False,
    ):
        y1, y0, c1, c0, match, n, x_name, c1_name, c0_name = _related_measures_args(y, x, c1, c0, match, ds, sub)
        if tail == 0:
            alternative = 'two-sided'
        elif tail == 1:
            alternative = 'greater'
        elif tail == -1:
            alternative = 'less'
        else:
            raise ValueError(f"tail={tail!r}")
        self.w, self.p = scipy.stats.wilcoxon(y1.x, y0.x, zero_method, correction, alternative)
        self._y = dataobj_repr(y1)
        self._x = x_name
        self._c1 = c1
        self._c0 = c0
        self.c1_mean = y1.mean()
        self.c0_mean = y0.mean()
        self.difference = y1 - y0
        self._match = dataobj_repr(match, True)
        self.tail = tail
        self.zero_method = zero_method
        self.correction = correction

    def __repr__(self):
        cmp = '=><'[self.tail]
        if self._match is None:
            desc = f"{self._c1} {cmp} {self._c0}"
        else:
            desc = f"{self._y} ~ {self._x}, {self._c1} {cmp} {self._c0}"
        return f"<{self.__class__.__name__}: {desc}; {self._asfmtext()}>"

    def _asfmtext(self, **_):
        return fmtxt.FMText([fmtxt.eq('W', self.w), ', ', fmtxt.peq(self.p)])


def pairwise(
        y: VarArg,
        x: CategorialArg,
        match: CategorialArg = None,
        sub: IndexArg = None,
        ds: Dataset = None,
        par: bool = True,
        corr: Union[None, str] = 'Hochberg',
        trend: Union[bool, str] = False,
        title: str = '{desc}',
        mirror: bool = False,
):
    """Pairwise comparison table

    Parameters
    ----------
    y : Var
        Dependent measure.
    x : categorial
        Categories to compare.
    match : None | Factor
        Repeated measures factor.
    sub : index-array
        Perform tests with a subset of the data.
    ds : Dataset
        If a Dataset is given, all data-objects can be specified as names of
        Dataset variables.
    par : bool
        Use parametric test for pairwise comparisons (use non-parametric
        tests if False).
    corr : None | 'hochberg' | 'bonferroni' | 'holm'
        Method for multiple comparison correction.
    trend
        Marker for a trend in pairwise comparisons.
    title
        Title for the table.
    mirror
        Redundant table including all row/column combinations.

    Returns
    -------
    table : FMText Table
        Table with results.
    """
    ct = Celltable(y, x, match=match, sub=sub, ds=ds, coercion=asvar)
    test = _pairwise(ct.get_data(), ct.all_within, par, corr, trend)

    # extract test results
    k = len(ct)
    indexes = test['pw_indexes']
    statistic = test['statistic']
    _K = test[statistic]
    _P = test['p']
    _Pc = mcp_adjust(_P, corr)
    _df = test['df']
    symbols = test['symbols']

    # create TABLE
    table = fmtxt.Table('l' + 'l' * (k - 1 + mirror))
    title_desc = "Pairwise {0}".format(test['test'])
    table.title(title.format(desc=title_desc))
    table.caption(test['caption'])

    # headings
    table.cell()
    cellnames = ct.cellnames()
    for name in cellnames[1 - mirror:]:
        table.cell(name)
    table.midrule()

    # tex_df = fmtxt.Element(df, "_", digits=0)
    if corr and not mirror:
        subrows = 3
    else:
        subrows = 2

    n_stars = 3 + bool(trend)
    for row in range(0, k - 1 + mirror):
        for subrow in range(subrows):  # contains t/p
            # names column
            if subrow == 0:
                table.cell(cellnames[row])
            else:
                table.cell()
            # rows
            for col in range(1 - mirror, k):
                if row == col:
                    table.cell()
                elif col > row:
                    index = indexes[(row, col)]
                    if subrow == 0:
                        table.cell(fmtxt.eq(statistic, _K[index], _df[index], stars=symbols[index], of=n_stars))
                    elif subrow == 1:
                        table.cell(fmtxt.peq(_P[index]))
                    elif subrow == 2:
                        table.cell(fmtxt.peq(_Pc[index], 'c'))
                elif mirror and corr and subrow == 0:
                    index = indexes[(col, row)]
                    table.cell(fmtxt.P(_Pc[index]))
                else:
                    table.cell()
    return table


def _pairwise(
        data: Sequence[Sequence[float]],  # list of groups/treatments
        within: bool,
        parametric: bool,
        corr: MCCArg,
        trend: Union[bool, str] = False,
        levels: Union[bool, dict] = True,
):
    """Pairwise tests

    Returns
    -------
    results : dict
        dictionary with results:
        'test': test name
        'caption': information about correction
        'statistic': abbreviation used for the staistic (e.g. 'Q')
        statistic: list of values
        'df': df
        'p': list of corresponding pa values
        'stars': list of n stars (ints)
        'pw_indexes': dict linking table index (i,j) to the list index for p etc.
    """
    # find test
    k = len(data)
    if k < 3:  # need no correction for single test
        corr = None
    if parametric:
        if within:
            test_func = scipy.stats.ttest_rel
            test_name = "T-Tests (paired samples)"
        else:
            test_func = scipy.stats.ttest_ind
            test_name = "T-Tests (independent samples)"
        statistic = "t"
    elif within:
        test_func = partial(scipy.stats.wilcoxon, alternative='two-sided')
        test_name = "Wilcoxon Signed-Rank Test"
        statistic = "z"
    else:
        test_func = partial(scipy.stats.mannwhitneyu, alternative='two-sided')
        test_name = "Mann-Whitney U Test"
        statistic = "u"

    # perform test
    _K = []  # kennwerte
    _P = []
    _df = []
    i = 0
    indexes = {}
    for x in range(k):
        for y in range(x + 1, k):
            Y1, Y2 = data[x], data[y]
            t, p = test_func(Y1, Y2)
            _K.append(t)
            if within:
                _df.append(len(Y1) - 1)
            else:
                _df.append(len(Y1) + len(Y2) - 2)

            _P.append(p)
            indexes[(x, y)] = indexes[(y, x)] = i
            i += 1
    # add stars
    if corr:
        p_adjusted = mcp_adjust(_P, corr)
    else:
        p_adjusted = _P
    # prepare output
    out = {'test': test_name,
           'caption': _get_correction_caption(corr, len(_P)),
           'statistic': statistic,
           statistic: _K,
           'df': _df,
           'p': _P,
           'stars': star(p_adjusted, int, levels, trend),
           'symbols': star(p_adjusted, str, levels, trend),
           'pw_indexes': indexes}
    return out


def pairwise_correlations(xs, sub=None, ds=None, labels={}):
    """Pairwise correlation table

    Parameters
    ----------
    xs : sequence of Var | NDVar
        Variables to correlate.
    sub : index
        Use only a subset of the data
    ds : Dataset
        If a Dataset is given, all data-objects can be specified as names of
        Dataset variables.
    labels : {str: str} dict
        Labels for ``xs`` in the table (mapping ``x.name`` to string labels).

    Returns
    -------
    pairwise_table : fmtxt.Table
        Table with pairwise correlations.
    """
    sub = assub(sub, ds)
    x_rows = [asnumeric(x, sub, ds) for x in xs]
    n_vars = len(x_rows)

    table = fmtxt.Table('l' + 'c' * n_vars)
    # header
    table.cell()
    for i in range(1, n_vars + 1):
        table.cell(i)
    table.midrule()
    # body
    for i_row, x_row in enumerate(x_rows):
        label = nice_label(x_row, labels)
        table.cell("%i  %s" % (i_row + 1, label))
        for x_col in x_rows:
            if x_col is x_row:
                table.cell()
            else:
                corr = Correlation(x_row, x_col)
                table.cell(fmtxt.stat(corr.r, drop0=True))
        table.endline()
    return table


def correlations(y, x, cat=None, sub=None, ds=None, asds=False):
    """Correlation with one or more predictors

    Parameters
    ----------
    y : Var
        First variable
    x : Var | list of Var
        second variable (or list of variables).
    cat : categorial
        Show correlations separately for different groups in the data.
    sub : index
        Use only a subset of the data
    ds : Dataset
        If a Dataset is given, all data-objects can be specified as names of
        Dataset variables.
    asds : bool
        Return correlations in Dataset instead of Table.

    Returns
    -------
    correlations : Table | Dataset
        Table or Dataset (if ``asds=True``) with correlations.
    """
    sub = assub(sub, ds)
    y = asvar(y, sub, ds)
    if isinstance(x, (Var, str)):
        x = (x,)
        print_x_name = False
    else:
        print_x_name = True
    x = [asvar(x_, sub, ds) for x_ in x]
    if cat is None:
        cat_cells = [None]
        n_cat = 1
    else:
        cat = ascategorial(cat, sub, ds)
        cat_cells = cat.cells
        n_cat = len(cat_cells)

    # correlation Dataset, nested:
    # x -> cat
    ds = Dataset()
    if print_x_name:
        ds['x'] = Factor([dataobj_repr(x_) for x_ in x], repeat=n_cat)

    if n_cat > 1:
        if isinstance(cat, Interaction):
            cat_names = [dataobj_repr(c) for c in cat.base]
            for i, name in enumerate(cat_names):
                ds[name] = Factor([cell[i] for cell in cat.cells], tile=len(x))
        elif isinstance(cat, Factor):
            cat_names = (dataobj_repr(cat),)
            ds[dataobj_repr(cat)] = Factor(cat.cells, tile=len(x))
        else:
            raise TypeError(repr(cat))
    else:
        cat_names = ()

    rs = []
    dfs = []
    ps = []
    for x_ in x:
        for cell in cat_cells:
            if cell is None:
                r, p, df = _corr(y.x, x_.x)
            else:
                sub = cat == cell
                r, p, df = _corr(y[sub].x, x_[sub].x)
            rs.append(r)
            dfs.append(df)
            ps.append(p)
    ds['r'] = Var(rs)
    ds['df'] = Var(dfs)
    p = Var(ps)
    ds['sig'] = star_factor(p)
    ds['p'] = p
    if asds:
        return ds

    table = fmtxt.Table('l' * (4 + print_x_name + len(cat_names)),
                        title="Correlations with %s" % dataobj_repr(y))
    if print_x_name:
        table.cell('x')
    table.cells(*cat_names)
    table.cells('r', 'df', '*' 'p')
    table.midrule()
    last_x = None
    for case in ds.itercases():
        if print_x_name:
            if case['x'] == last_x:
                table.cell('')
            else:
                table.cell(case['x'])
                last_x = case['x']
        for name in cat_names:
            table.cell(case[name])
        table.cell(fmtxt.stat(case['r'], '%.3f', drop0=True))
        table.cell(case['df'])
        table.cell(case['sig'])
        table.cell(fmtxt.p(case['p']))
    return table


def _corr(y, x):
    n = len(y)
    assert len(x) == n
    df = n - 2
    r = np.corrcoef(y, x)[0, 1]
    if r == 1:
        return r, 0., df
    t = r / np.sqrt((1 - r ** 2) / df)
    p = scipy.stats.t.sf(np.abs(t), df) * 2
    return r, p, df


class bootstrap_pairwise:
    def __init__(self, y, x, match=None, sub=None,
                 samples=1000, replacement=True,
                 title="Bootstrapped Pairwise Tests", ds=None):
        sub = assub(sub, ds)
        y = asvar(y, sub, ds)
        x = asfactor(x, sub, ds)
        assert len(y) == len(x), "data length mismatch"
        if match is not None:
            match = ascategorial(match, sub, ds)
            assert len(match) == len(y), "data length mismatch"

        # prepare data container
        resampled = np.empty((samples + 1, len(y)))  # sample x subject within category
        resampled[0] = y.x
        # fill resampled
        for i, y_i in enumerate(resample(y, samples, replacement, match), 1):
            resampled[i] = y_i.x
        self.resampled = resampled

        cells = x.cells
        n_groups = len(cells)

        if match is not None:
            # if there are several values per x%match cell, take the average
            # T: indexes to transform y.x to [x%match, value]-array
            match_cell_ids = match.cells
            group_size = len(match_cell_ids)
            T = None
            i = 0
            for x_cell in cells:
                for match_cell in match_cell_ids:
                    source_indexes = np.where((x == x_cell) * (match == match_cell))[0]
                    if T is None:
                        n_cells = n_groups * group_size
                        T = np.empty((n_cells, len(source_indexes)), dtype=int)
                    T[i, :] = source_indexes
                    i += 1

            if T.shape[1] == 1:
                T = T[:, 0]
                ordered = resampled[:, T]
            else:
                ordered = resampled[:, T].mean(axis=2)
            self.ordered = ordered

            # t-tests
            n_comparisons = sum(range(n_groups))
            t = np.empty((samples + 1, n_comparisons))
            comp_names = []
            one_group = np.arange(group_size)
            groups = [one_group + i * group_size for i in range(n_groups)]
            for i, (g1, g2) in enumerate(itertools.combinations(range(n_groups), 2)):
                group_1 = groups[g1]
                group_2 = groups[g2]
                diffs = ordered[:, group_1] - ordered[:, group_2]
                t[:, i] = (np.mean(diffs, axis=1) * np.sqrt(group_size) /
                           np.std(diffs, axis=1, ddof=1))
                comp_names.append(' - '.join((cells[g1], cells[g2])))

            self.diffs = diffs
            self.t_resampled = np.max(np.abs(t[1:]), axis=1)
            self.t = t = t[0]
        else:
            raise NotImplementedError

        self._Y = y
        self._X = x
        self._group_names = cells
        self._group_data = np.array([ordered[0, g] for g in groups])
        self._group_size = group_size
        self._df = group_size - 1
        self._match = match
        self._n_samples = samples
        self._replacement = replacement
        self._comp_names = comp_names
        self._p_parametric = self.test_param(t)
        self._p_boot = self.test_boot(t)
        self.title = title

    def __repr__(self):
        out = ['bootstrap_pairwise(', self._Y.name, self._X.name]
        if self._match:
            out.append('match=%s ' % self._match.name)
        out.append('saples=%i ' % self._n_samples)
        out.append('replacement=%s)' % self._replacement)
        return ''.join(out)

    def __str__(self):
        return str(self.table())

    def table(self):
        table = fmtxt.Table('lrrrr')
        table.title(self.title)
        table.caption("Results based on %i samples" % self._n_samples)
        table.cell('Comparison')
        table.cell(fmtxt.symbol('t', self._df))
        table.cell(fmtxt.symbol('p', 'param'))
        table.cell(fmtxt.symbol('p', 'corr'))
        table.cell(fmtxt.symbol('p', 'boot'))
        table.midrule()

        p_corr = mcp_adjust(self._p_parametric)
        stars_parametric = star(p_corr)
        stars_boot = star(self._p_boot)

        for name, t, p1, pc, s1, p2, s2 in zip(self._comp_names, self.t,
                                                self._p_parametric, p_corr,
                                                stars_parametric,
                                                self._p_boot, stars_boot):
            table.cell(name)
            table.cell(t, fmt='%.2f')
            table.cell(fmtxt.p(p1))
            table.cell(fmtxt.p(pc, stars=s1))
            table.cell(fmtxt.p(p2, stars=s2))
        return table

    def plot_t_dist(self):
        # http://stackoverflow.com/questions/4150171/how-to-create-a-density-plot-in-matplotlib
        from matplotlib import pyplot

        t = self.t_resampled
        density = scipy.stats.gaussian_kde(t)
#        density.covariance_factor = lambda : .25
#        density._compute_covariance()
        xs = np.linspace(0, max(t), 200)
        pyplot.plot(xs, density(xs))

    def plot_dv_dist(self):
        from matplotlib import pyplot

        xs = np.linspace(np.min(self._group_data), np.max(self._group_data), 200)
        for i, name in enumerate(self._group_names):
            density = scipy.stats.gaussian_kde(self._group_data[i])
            pyplot.plot(xs, density(xs), label=name)
        pyplot.legend()

    def test_param(self, t):
        return scipy.stats.t.sf(np.abs(t), self._group_size - 1) * 2

    def test_boot(self, t):
        "t: scalar or array; returns p for each t"
        test = self.t_resampled[:, None] > np.abs(t)
        return np.sum(test, axis=0) / self.t_resampled.shape[0]



"""Plot source space NDVars with mayavi/pysurfer."""
# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>

from ._brain import (
    SequencePlotter,
    brain,
    cluster, dspm, p_map, surfer_brain, annot, annot_legend,
    butterfly,
    bin_table, dspm_bin_table,
    copy,
)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
import os


def rename_mri(subject_from, subject_to, subjects_dir, preview=False):
    """Rename an MRI subject

    Parameters
    ----------
    subject_from : str
        Name of the subject that should be renamed.
    subject_to : str
        Name of the new subject name.
    subjects_dir : str
        Subjects-directory.
    preview : bool
        Show all the files that would be renamed instead of actually renaming
        them.

    Notes
    -----
    Does not update the ``MRI scaling parameters.cfg`` file for scaled MRIs.
    """
    sdir = os.path.join(subjects_dir, subject_from)
    if not os.path.exists(sdir):
        raise IOError("Directory for %s does not exist at %s" %
                      (subject_from, subjects_dir))

    n_sdir = len(subjects_dir) + 1
    n_from = len(subject_from)

    pairs = []
    for root, dirs, files in os.walk(sdir):
        for filename in files:
            if filename.startswith(subject_from):
                old = os.path.join(root, filename)
                new = os.path.join(root, subject_to + filename[n_from:])
                pairs.append((old, new))
    pairs.append((sdir, os.path.join(subjects_dir, subject_to)))

    if preview:
        for old, new in pairs:
            print("  %s\n->%s" % (old[n_sdir:], new[n_sdir:]))
    else:
        for old, new in pairs:
            os.rename(old, new)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from math import floor, log10
from os import makedirs
from os.path import basename, dirname, exists, expanduser, isdir, join

import matplotlib as mpl
import numpy as np

from . import fmtxt, plot, testnd
from .plot._base import POINT
from ._data_obj import combine


# usage:  with mpl.rc_context(RC):
FONT = 'Helvetica'
RC = {
    'figure.dpi': 300,
    'savefig.dpi': 300,
    'savefig.transparent': True,
    # Font
    'font.family': 'sans-serif',
    'font.sans-serif': FONT,
    'font.size': 9,
    # make sure equations use same font
    'mathtext.fontset': 'custom',
    'font.cursive': FONT,
    'font.serif': FONT,
    # subplot
    'figure.subplot.top': 0.95,
    # legend
    'legend.fontsize': 6,
    'legend.frameon': False,
}
for key in mpl.rcParams:
    if 'width' in key:
        RC[key] = mpl.rcParams[key] * 0.5


class PlotDestDir:
    """Generate paths for saving plots in figure-specific subdirectories

    Parameters
    ----------
    root : str
        Directory in which to save files.
    pix_fmt : str
        Pixel graphics format (default ``png``).
    vec_fmt : str
        Vector graphics format (default ``pdf``).
    name : str
        Name for the info report (default is ``basename(root)``).
    """
    def __init__(self, root, pix_fmt='png', vec_fmt='pdf', name=None):
        root = expanduser(root)
        if not exists(root):
            makedirs(root)
        else:
            assert isdir(root)
        assert pix_fmt.isalnum()
        assert vec_fmt.isalnum()
        if name is None:
            name = basename(root)
            if not name:
                name = basename(dirname(root))
        self.root = root
        self._pix_fmt = pix_fmt
        self._vec_fmt = vec_fmt
        self.pix = join(root, '%s.' + pix_fmt)
        self.vec = join(root, '%s.' + vec_fmt)
        self.mov = join(root, '%s.mov')
        self.txt = join(root, '%s.txt')
        self.name = name
        self.report = fmtxt.Report(name)
        self._active_section = [self.report]

    def with_ext(self, ext):
        """Generate path template ``%s.{ext}``"""
        assert ext.isalnum()
        return join(self.root, '%s.' + ext)

    def subdir(self, dirname, name=None):
        """PlotDestDir object for a sub-directory"""
        return PlotDestDir(join(self.root, dirname), self._pix_fmt, self._vec_fmt, name)

    # MARK:  report

    def section(self, heading, level=1):
        if level <= 0:
            raise ValueError("level=%r; must be >= 1, section 0 is the document")
        elif level > len(self._active_section):
            raise RuntimeError("Can't add section with level %i before adding "
                               "section with level %i" % (level, level - 1))
        while len(self._active_section) > level:
            self._active_section.pop(-1)
        section = self._active_section[-1].add_section(heading)
        self._active_section.append(section)

    def info(self, content):
        """Add ``info_string`` to the info list"""
        section = self._active_section[-1]
        section.append(content)

    def save_info(self, format='html'):
        """Save info to ``info.txt``"""
        dst = join(self.root, self.name)
        try:
            getattr(self.report, 'save_' + format)(dst)
        except AttributeError:
            raise ValueError("format=%r; Invalid format" % (format,))


def cname(cid):
    if isinstance(cid, tuple):
        return '-'.join(map(str, cid))
    else:
        return str(cid)


class ClusterPlotter:
    """Make plots for spatio-temporal clusters

    returned by :meth:`MneExperiment.load_result_plotter`

    Parameters
    ----------
    ds : Dataset
        Dataset with the data on which the test is based.
    res : NDTest
        Test result object with spatio-temporal cluster test result.
    colors : dict
        Colors for plotting data in a ``{cell: color}`` dictionary.
    dst : str
        Directory in which to place results.
    vec_fmt : str
        Format for vector graphics (default 'pdf').
    pix_fmt : str
        Format for pixel graphics (default 'png').
    labels : dict
        Labels for data in a ``{cell: label}`` dictionary (the default is to
        use cell names).
    h : scalar
        Plot height in inches (default 1.2).
    rc : dict
        Matplotlib rc-parameters dictionary (the default is optimized for the
        default plot size ``h=1.2``).

    Notes
    -----
    After loading a :class:`ClusterPlotter`, its ``rc``, ``colors``, ``labels``
    and ``h`` attributes can be updated to create different plot layouts without
    reloading the data.
    """
    def __init__(self, ds, res, colors, dst, vec_fmt='pdf', pix_fmt='png',
                 labels=None, h=1.2, rc=None):
        self.rc = RC.copy()
        if rc is not None:
            self.rc.update(rc)
        self.ds = ds
        self.res = res
        self.colors = colors
        self.labels = labels
        self.h = h
        self._dst = PlotDestDir(dst, pix_fmt, vec_fmt)
        self._is_anova = isinstance(self.res, testnd.ANOVA)

    def _ids(self, ids):
        if isinstance(ids, (float, int)):
            return self._ids_for_p(ids)
        elif isinstance(ids, dict):
            if not self._is_anova:
                raise TypeError("ids can not be dict for results other than ANOVA")
            out = []
            for effect, cids in ids.items():
                if isinstance(cids, float):
                    out.extend(self._ids_for_p(cids, effect))
                else:
                    out.extend((effect, cid) for cid in cids)
            return out
        else:
            return ids

    def _ids_for_p(self, p, effect=None):
        "Find cluster IDs for clusters with p-value <= p"
        if effect is None:
            clusters = self.res.find_clusters(p)
        else:
            clusters = self.res.find_clusters(p, effect=effect)
            clusters[:, 'effect'] = effect

        if self._is_anova:
            return list(zip(clusters['effect'], clusters['id']))
        else:
            return clusters['id']

    def _get_clusters(self, ids):
        return [self._get_cluster(cid) for cid in ids]

    def _get_cluster(self, cid):
        if self._is_anova:
            effect, cid = cid
            return self.res.cluster(cid, effect)
        else:
            return self.res.cluster(cid)

    def plot_color_list(self, name, cells, w=None, colors=None):
        if colors is None:
            colors = self.colors

        with mpl.rc_context(self.rc):
            p = plot.ColorList(colors, cells, self.labels, w=w, show=False)
            p.save(self._dst.vec % "colorlist %s" % name, transparent=True)
            p.close()

    def plot_color_grid(self, name, row_cells, column_cells):
        with mpl.rc_context(self.rc):
            p = plot.ColorGrid(row_cells, column_cells, self.colors, labels=self.labels)
            p.save(self._dst.vec % "colorgrid %s" % name, transparent=True)
            p.close()

    def plot_clusters_spatial(self, ids, views, w=600, h=480, prefix=''):
        """Plot spatial extent of the clusters

        Parameters
        ----------
        ids : sequence | dict | scalar <= 1
            IDs of the clusters that should be plotted. For ANOVA results, this
            should be an ``{effect_name: id_list}`` dict. Instead of a list of
            IDs a scalar can be provided to plot all clusters with p-values
            smaller than this.
        views : str | list of str | dict
            Can a str or list of str to use the same views for all clusters. A dict
            can have as keys labels or cluster IDs.
        w, h : int
            Size in pixels. The default (600 x 480) corresponds to 2 x 1.6 in
            at 300 dpi.
        prefix : str
            Prefix to use for the image files (optional, can be used to
            distinguish different groups of images sharing the same color-bars).

        Notes
        -----
        The horizontal colorbar is 1.5 in wide, the vertical colorbar is 1.6 in
        high.
        """
        ids = self._ids(ids)
        clusters = self._get_clusters(ids)
        clusters_spatial = [c.sum('time') for c in clusters]
        if isinstance(views, str):
            views = (views,)

        # vmax
        vmin = min(c.min() for c in clusters_spatial)
        vmax = max(c.max() for c in clusters_spatial)
        abs_vmax = max(vmax, abs(vmin))

        # anatomical extent
        brain_colorbar_done = False
        for cid, cluster in zip(ids, clusters_spatial):
            name = cname(cid)
            if prefix:
                name = prefix + ' ' + name

            for hemi in ('lh', 'rh'):
                if not cluster.sub(source=hemi).any():
                    continue
                brain = plot.brain.cluster(cluster, abs_vmax, views='lat',
                                           background=(1, 1, 1), colorbar=False,
                                           parallel=True, hemi=hemi, w=w, h=h)
                for view in views:
                    brain.show_view(view)
                    brain.save_image(self._dst_pix % ' '.join((name, hemi, view)),
                                     'rgba', True)

                if not brain_colorbar_done:
                    with mpl.rc_context(self.rc):
                        label = "Sum of %s-values" % cluster.info['meas']
                        clipmin = 0 if vmin == 0 else None
                        clipmax = 0 if vmax == 0 else None
                        if prefix:
                            cbar_name = '%s cbar %%s' % prefix
                        else:
                            cbar_name = 'cbar %s'

                        h_cmap = 0.7 + POINT * mpl.rcParams['font.size']
                        p = brain.plot_colorbar(label, clipmin=clipmin, clipmax=clipmax,
                                                width=0.1, h=h_cmap, w=1.5, show=False)
                        p.save(self._dst.vec % cbar_name % 'h', transparent=True)
                        p.close()

                        w_cmap = 0.8 + 0.1 * abs(floor(log10(vmax)))
                        p = brain.plot_colorbar(label, clipmin=clipmin, clipmax=clipmax,
                                                width=0.1, h=1.6, w=w_cmap,
                                                orientation='vertical', show=False)
                        p.save(self._dst.vec % cbar_name % 'v', transparent=True)
                        p.close()

                        brain_colorbar_done = True

                brain.close()

    def _get_data(self, model, sub, subagg):
        """Plot values in cluster

        Parameters
        ----------
        subagg : str
           Index in ds: within index, collapse across other predictors.
        """
        ds = self.ds
        modelname = model

        if sub:
            ds = ds.sub(sub)
            modelname += '[%s]' % sub

        if subagg:
            idx_subagg = ds.eval(subagg)
            ds_full = ds.sub(np.invert(idx_subagg))
            ds_agg = ds.sub(idx_subagg).aggregate("subject", drop_bad=True)
            ds = combine((ds_full, ds_agg), incomplete='fill in')
            ds['condition'] = ds.eval(model).as_factor()
            model = 'condition'
            modelname += '(agg %s)' % subagg

        return ds, model, modelname

    def plot_values(self, ids, model, ymax, ymin, dpi=300, sub=None,
                    subagg=None, cells=None, pairwise=False, colors=None,
                    prefix=None, w=None, filter=None, legend=False):
        """Plot values in cluster

        Parameters
        ----------
        ids : sequence | dict | scalar <= 1
            IDs of the clusters that should be plotted. For ANOVA results, this
            should be an ``{effect_name: id_list}`` dict. Instead of a list of
            IDs a scalar can be provided to plot all clusters with p-values
            smaller than this.
        model : str
            Model defining cells which to plot separately.
        ymax : scalar
            Top of the y-axis.
        ymin : scalar
            Bottom of the y axis.
        dpi : int
            Figure DPI.
        sub : str
            Only use a subset of the data.
        subagg : str
           Index in ds: within index, collapse across other predictors.
        cells : sequence of cells in model
            Modify visible cells and their order. Only applies to the barplot.
            Does not affect filename.
        pairwise : bool
            Add pairwise tests to barplots.
        colors : dict
            Substitute colors (default are the colors provided at
            initialization).
        prefix : str
            Prefix to use for the image files (optional, can be used to
            distinguish different groups of images sharing the same color-bars).
        w : scalar
            UTS-stat plot width (default is ``2 * h``).
        filter : Filter
            Filter signal for display purposes (optional).
        legend : bool
            Plot a color legend.
        """
        if w is None:
            w = self.h * 2
        ds, model, modelname = self._get_data(model, sub, subagg)
        ids = self._ids(ids)
        if colors is None:
            colors = self.colors

        src = ds['srcm']
        n_cells = len(ds.eval(model).cells)
        w_bar = (n_cells * 2 + 4) * (self.h / 12)
        with mpl.rc_context(self.rc):
            for cid in ids:
                name = cname(cid)
                if prefix:
                    name = prefix + ' ' + name
                cluster = self._get_cluster(cid)
                y_mean = src.mean(cluster != 0)
                y_tc = src.mean(cluster.any('time'))

                # barplot
                p = plot.Barplot(
                    y_mean, model, 'subject', None, cells, pairwise, ds=ds,
                    trend=False, corr=None, title=None, frame=False,
                    yaxis=False, ylabel=False, colors=colors, bottom=ymin,
                    top=ymax, w=w_bar, h=self.h, xlabel=None, xticks=None,
                    tight=False, test_markers=False, show=False)
                p.save(self._dst.vec % ' '.join((name, modelname, 'barplot')),
                       dpi=dpi, transparent=True)
                p.close()

                # time-course
                if filter is not None:
                    y_tc = filter.filtfilt(y_tc)
                p = plot.UTSStat(
                    y_tc, model, match='subject', ds=ds, error='sem',
                    colors=colors, title=None, axtitle=None, frame=False,
                    bottom=ymin, top=ymax, legend=None, ylabel=None,
                    xlabel=None, w=w, h=self.h, tight=False, show=False)
                dt = y_tc.time.tstep / 2.
                mark_start = cluster.info['tstart'] - dt
                mark_stop = cluster.info['tstop'] - dt
                p.add_vspan(mark_start, mark_stop, color='k', alpha=0.1, zorder=-2)
                p.save(self._dst.vec % ' '.join((name, modelname, 'timecourse')),
                       dpi=dpi, transparent=True)
                p.close()

                # legend (only once)
                if legend:
                    p.save_legend(self._dst.vec % (modelname + ' legend'),
                                  transparent=True)
                    legend = False



# when installed through setup.py, the first line is replaced with the normal,
# non-framework build python. Replace bin/python after ENV with::
# #!/Users/christian/anaconda/envs/ENV/python.app/Contents/MacOS/python
from argparse import ArgumentParser
import os
import sys
from warnings import catch_warnings, filterwarnings

from traits.trait_base import ETSConfig
from IPython import start_ipython
from IPython.paths import get_ipython_dir
with catch_warnings():
    filterwarnings('ignore', "can't resolve package from __spec__", ImportWarning)
    import eelbrain


parser = ArgumentParser('eelbrain', description='Start iPython with eelbrain profile')
parser.add_argument('--version', action='version', version=eelbrain.__version__)
parser.add_argument('--debug-warnings', action='store_true', help='Print tracebacks for warnings')
parser.add_argument('--show-warnings', action='store_true', help='Show all warnings; by default, many warnings that are probably irrelevant are suppressed')


if __name__ == '__main__':
    args = parser.parse_args()
    del sys.argv[1:]
    if args.debug_warnings:
        from eelbrain._utils.debug import warn_with_traceback
        warn_with_traceback()
    if args.show_warnings:
        eelbrain._config.SUPPRESS_WARNINGS = False

    # set ETS toolkit before it is launched
    ETSConfig.toolkit = 'wx'

    if not os.path.exists(os.path.join(get_ipython_dir(), 'profile_eelbrain')):
        from IPython.core.profileapp import ProfileCreate
        app = ProfileCreate()
        app.parse_command_line(('eelbrain', '--quiet'))
        app.init_config_files()
        file_path = os.path.join(get_ipython_dir(), 'profile_eelbrain',
                                 'startup', 'import_eelbrain.py')
        with open(file_path, 'w') as fid:
            fid.write("from eelbrain import *\n")

    sys.exit(start_ipython(profile='eelbrain'))
